import re
import os
import numpy as np
import itertools
import collections
from unidecode import unidecode
from .._utils._utils import download_file
from ._tatabahasa import stopword_tatabahasa, stopwords, stopwords_calon
from ._english_words import _english_words
from ._malay_words import _malay_words
from .._transformer._xlnet_model.prepro_utils import (
    preprocess_text,
    encode_ids,
    encode_pieces,
)
from .. import home
import json

STOPWORDS = set(stopwords + stopword_tatabahasa + stopwords_calon)
STOPWORD_CALON = set(stopwords_calon)
VOWELS = 'aeiou'
PHONES = ['sh', 'ch', 'ph', 'sz', 'cz', 'sch', 'rz', 'dz']
ENGLISH_WORDS = _english_words
MALAY_WORDS = _malay_words


class SentencePieceTokenizer:
    def __init__(self, v, sp_model):
        self.vocab = v
        self.sp_model = sp_model

    def tokenize(self, string):
        return encode_pieces(
            self.sp_model, string, return_unicode = False, sample = False
        )

    def convert_tokens_to_ids(self, tokens):
        return [self.sp_model.PieceToId(piece) for piece in tokens]

    def convert_ids_to_tokens(self, ids):
        return [self.sp_model.IdToPiece(i) for i in ids]


def _isWord(word):
    if word:
        consecutiveVowels = 0
        consecutiveConsonents = 0
        for idx, letter in enumerate(word.lower()):
            vowel = True if letter in VOWELS else False
            if idx:
                prev = word[idx - 1]
                prevVowel = True if prev in VOWELS else False
                if not vowel and letter == 'y' and not prevVowel:
                    vowel = True
                if prevVowel != vowel:
                    consecutiveVowels = 0
                    consecutiveConsonents = 0
            if vowel:
                consecutiveVowels += 1
            else:
                consecutiveConsonents += 1
            if consecutiveVowels >= 3 or consecutiveConsonents > 3:
                return False
            if consecutiveConsonents == 3:
                subStr = word[idx - 2 : idx + 1]
                if any(phone in subStr for phone in PHONES):
                    consecutiveConsonents -= 1
                    continue
                return False
    return True


_list_laughing = {
    'huhu',
    'haha',
    'gaga',
    'hihi',
    'wkawka',
    'wkwk',
    'kiki',
    'keke',
    'huehue',
}


def remove_links_alias(string):
    string = unidecode(string)
    string = re.sub(
        r'\w+:\/{2}[\d\w-]+(\.[\d\w-]+)*(?:(?:\/[^\s/]*))*', '', string
    )
    string = re.sub(r'[ ]+', ' ', string).strip().split()
    string = [w for w in string if w[0] != '@']
    string = [w.title() if w[0].isupper() else w for w in string]
    return ' '.join(string)


def malaya_textcleaning(string):
    """
    use by normalizer, spell
    remove links, hashtags, alias
    only accept A-Z, a-z
    remove any laugh
    remove any repeated char more than 2 times
    remove most of nonsense words
    """
    string = re.sub(
        'http\S+|www.\S+',
        '',
        ' '.join(
            [
                word
                for word in string.split()
                if word.find('#') < 0 and word.find('@') < 0
            ]
        ),
    )
    string = unidecode(string).replace('.', '. ').replace(',', ' , ')
    string = re.sub('[^\'"A-Za-z\- ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string.lower()).strip()
    string = [word for word in string.lower().split() if _isWord(word)]
    string = [
        word
        for word in string
        if not any([laugh in word for laugh in _list_laughing])
        and word[: len(word) // 2] != word[len(word) // 2 :]
    ]
    string = ' '.join(string)
    string = (
        ''.join(''.join(s)[:2] for _, s in itertools.groupby(string))
    ).split()
    return ' '.join([word for word in string if word not in STOPWORDS])


def normalizer_textcleaning(string):
    """
    use by normalizer, spell
    remove links, hashtags, alias
    only accept A-Z, a-z
    remove any laugh
    remove any repeated char more than 2 times
    """
    string = re.sub(
        'http\S+|www.\S+',
        '',
        ' '.join(
            [
                word
                for word in string.split()
                if word.find('#') < 0 and word.find('@') < 0
            ]
        ),
    )
    string = re.sub('[^A-Za-z ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string).strip()
    string = [
        word.title() if word.isupper() else word
        for word in string.split()
        if len(word)
    ]
    string = [
        word
        for word in string
        if not any([laugh in word for laugh in _list_laughing])
    ]
    string = ' '.join(string)
    return ''.join(''.join(s)[:2] for _, s in itertools.groupby(string))


def simple_textcleaning(string, lowering = True):
    """
    use by topic modelling
    only accept A-Z, a-z
    """
    string = unidecode(string)
    string = re.sub('[^A-Za-z ]+', ' ', string)
    return re.sub(r'[ ]+', ' ', string.lower() if lowering else string).strip()


def entities_textcleaning(string, lowering = True):
    """
    use by entities recognition, pos recognition and dependency parsing
    """
    string = re.sub('[^A-Za-z0-9\-() ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string).strip()
    original_string = string.split()
    if lowering:
        string = string.lower()
    string = [
        (original_string[no], word.title() if word.isupper() else word)
        for no, word in enumerate(string.split())
        if len(word)
    ]
    return [s[0] for s in string], [s[1] for s in string]


def summary_textcleaning(string):
    original_string = string
    string = re.sub('[^A-Za-z0-9\-\/\'"\.\, ]+', ' ', unidecode(string))
    return original_string, re.sub(r'[ ]+', ' ', string.lower()).strip()


def get_hashtags(string):
    return [hash.lower() for hash in re.findall('#(\w+)', string)]


def split_by_dot(string):
    string = re.sub(
        r'(?<!\d)\.(?!\d)',
        'SPLITTT',
        string.replace('\n', '').replace('/', ' '),
    )
    string = string.split('SPLITTT')
    return [re.sub(r'[ ]+', ' ', sentence).strip() for sentence in string]


def language_detection_textcleaning(string):
    string = re.sub(
        u'[0-9!@#$%^&*()_\-+{}|\~`\'";:?/.>,<]',
        ' ',
        string.lower(),
        flags = re.UNICODE,
    )
    return re.sub(r'[ ]+', ' ', string).strip()


def pos_entities_textcleaning(string):
    """
    use by text entities and pos
    remove links, hashtags, alias
    """
    string = re.sub(
        'http\S+|www.\S+',
        '',
        ' '.join(
            [i for i in string.split() if i.find('#') < 0 and i.find('@') < 0]
        ),
    )
    string = unidecode(string).replace('.', ' . ').replace(',', ' , ')
    string = re.sub('[^A-Za-z\- ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string).strip()
    return ' '.join(
        [
            word.title() if word.isupper() else word
            for word in string.split()
            if len(word)
        ]
    )


def classification_textcleaning(string, no_stopwords = False, lowering = True):
    """
    stemmer, summarization, topic-modelling
    remove links, hashtags, alias
    """
    string = re.sub(
        'http\S+|www.\S+',
        '',
        ' '.join(
            [i for i in string.split() if i.find('#') < 0 and i.find('@') < 0]
        ),
    )
    string = unidecode(string).replace('.', ' . ').replace(',', ' , ')
    string = re.sub('[^A-Za-z ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string).strip()
    if no_stopwords:
        string = ' '.join(
            [
                i
                for i in re.findall('[\\w\']+|[;:\-\(\)&.,!?"]', string)
                if len(i)
            ]
        )
    else:
        string = ' '.join(
            [
                i
                for i in re.findall('[\\w\']+|[;:\-\(\)&.,!?"]', string)
                if len(i) and i not in STOPWORDS
            ]
        )
    if lowering:
        return string.lower()
    else:
        return ' '.join(
            [
                word.title() if word.isupper() else word
                for word in string.split()
                if len(word)
            ]
        )


def separate_dataset(trainset):
    datastring = []
    datatarget = []
    for i in range(len(trainset.data)):
        data_ = trainset.data[i].split('\n')
        data_ = list(filter(None, data_))
        datastring += data_
        for n in range(len(data_)):
            datatarget.append(trainset.target[i])
    return datastring, datatarget


def print_topics_modelling(
    topics, feature_names, sorting, n_words = 20, return_df = True
):
    if return_df:
        try:
            import pandas as pd
        except:
            raise Exception(
                'pandas not installed. Please install it and try again or set `return_df = False`'
            )
    df = {}
    for i in range(topics):
        words = []
        for k in range(n_words):
            words.append(feature_names[sorting[i, k]])
        df['topic %d' % (i)] = words
    if return_df:
        return pd.DataFrame.from_dict(df)
    else:
        return df


def str_idx(corpus, dic, maxlen, UNK = 0):
    X = np.zeros((len(corpus), maxlen))
    for i in range(len(corpus)):
        for no, k in enumerate(corpus[i].split()[:maxlen][::-1]):
            X[i, -1 - no] = dic.get(k, UNK)
    return X


def stemmer_str_idx(corpus, dic, UNK = 3):
    X = []
    for i in corpus:
        ints = []
        for k in i:
            ints.append(dic.get(k, UNK))
        X.append(ints)
    return X


def pad_sentence_batch(sentence_batch, pad_int):
    padded_seqs = []
    seq_lens = []
    max_sentence_len = max([len(sentence) for sentence in sentence_batch])
    for sentence in sentence_batch:
        padded_seqs.append(
            sentence + [pad_int] * (max_sentence_len - len(sentence))
        )
        seq_lens.append(len(sentence))
    return padded_seqs, seq_lens


def char_str_idx(corpus, dic, UNK = 2):
    maxlen = max([len(i) for i in corpus])
    X = np.zeros((len(corpus), maxlen))
    for i in range(len(corpus)):
        for no, k in enumerate(corpus[i][:maxlen]):
            X[i, no] = dic.get(k, UNK)
    return X


def generate_char_seq(batch, dic, UNK = 2):
    maxlen_c = max([len(k) for k in batch])
    x = [[len(i) for i in k] for k in batch]
    maxlen = max([j for i in x for j in i])
    temp = np.zeros((len(batch), maxlen_c, maxlen), dtype = np.int32)
    for i in range(len(batch)):
        for k in range(len(batch[i])):
            for no, c in enumerate(batch[i][k][::-1]):
                temp[i, k, -1 - no] = dic.get(c, UNK)
    return temp


def build_dataset(words, n_words, included_prefix = True):
    count = (
        [['GO', 0], ['PAD', 1], ['EOS', 2], ['UNK', 3]]
        if included_prefix
        else []
    )
    count.extend(collections.Counter(words).most_common(n_words))
    dictionary = dict()
    for word, _ in count:
        dictionary[word] = len(dictionary)
    data = list()
    for word in words:
        index = dictionary.get(word, 3)
        data.append(index)
    reversed_dictionary = dict(zip(dictionary.values(), dictionary.keys()))
    return data, count, dictionary, reversed_dictionary


def multireplace(string, replacements):
    substrs = sorted(replacements, key = len, reverse = True)
    regexp = re.compile('|'.join(map(re.escape, substrs)))
    return regexp.sub(lambda match: replacements[match.group(0)], string)


alphabets = '([A-Za-z])'
prefixes = (
    '(Mr|St|Mrs|Ms|Dr|Prof|Capt|Cpt|Lt|Mt|Puan|puan|Tuan|tuan|sir|Sir)[.]'
)
suffixes = '(Inc|Ltd|Jr|Sr|Co)'
starters = '(Mr|Mrs|Ms|Dr|He\s|She\s|It\s|They\s|Their\s|Our\s|We\s|But\s|However\s|That\s|This\s|Wherever|Dia|Mereka|Tetapi|Kita|Itu|Ini|Dan|Kami)'
acronyms = '([A-Z][.][A-Z][.](?:[A-Z][.])?)'
websites = '[.](com|net|org|io|gov|me|edu|my)'
another_websites = '(www|http|https)[.]'
digits = '([0-9])'


def split_into_sentences(text):
    text = unidecode(text)
    text = ' ' + text + '  '
    text = text.replace('\n', ' ')
    text = re.sub(prefixes, '\\1<prd>', text)
    text = re.sub(websites, '<prd>\\1', text)
    text = re.sub(another_websites, '\\1<prd>', text)
    if '...' in text:
        text = text.replace('...', '<prd><prd><prd>')
    if 'Ph.D' in text:
        text = text.replace('Ph.D.', 'Ph<prd>D<prd>')
    text = re.sub('\s' + alphabets + '[.] ', ' \\1<prd> ', text)
    text = re.sub(acronyms + ' ' + starters, '\\1<stop> \\2', text)
    text = re.sub(
        alphabets + '[.]' + alphabets + '[.]' + alphabets + '[.]',
        '\\1<prd>\\2<prd>\\3<prd>',
        text,
    )
    text = re.sub(
        alphabets + '[.]' + alphabets + '[.]', '\\1<prd>\\2<prd>', text
    )
    text = re.sub(' ' + suffixes + '[.] ' + starters, ' \\1<stop> \\2', text)
    text = re.sub(' ' + suffixes + '[.]', ' \\1<prd>', text)
    text = re.sub(' ' + alphabets + '[.]', ' \\1<prd>', text)
    text = re.sub(digits + '[.]' + digits, '\\1<prd>\\2', text)
    if '”' in text:
        text = text.replace('.”', '”.')
    if '"' in text:
        text = text.replace('."', '".')
    if '!' in text:
        text = text.replace('!"', '"!')
    if '?' in text:
        text = text.replace('?"', '"?')
    text = text.replace('.', '.<stop>')
    text = text.replace('?', '?<stop>')
    text = text.replace('!', '!<stop>')
    text = text.replace('<prd>', '.')
    sentences = text.split('<stop>')
    sentences = sentences[:-1]
    sentences = [s.strip() for s in sentences if len(s) > 10]
    return sentences


def end_of_chunk(prev_tag, tag):
    if not len(prev_tag):
        return False
    if prev_tag != tag:
        return True


def start_of_chunk(prev_tag, tag):
    if not len(prev_tag):
        return True
    if prev_tag != tag:
        return False


def tag_chunk(seq):
    words = [i[0] for i in seq]
    seq = [i[1] for i in seq]
    prev_tag = ''
    begin_offset = 0
    chunks = []
    for i, chunk in enumerate(seq):
        if end_of_chunk(prev_tag, chunk):
            chunks.append((prev_tag, begin_offset, i - 1))
            prev_tag = ''
        if start_of_chunk(prev_tag, chunk):
            begin_offset = i
        prev_tag = chunk
    res = {'words': words, 'tags': []}
    for chunk_type, chunk_start, chunk_end in chunks:
        tag = {
            'text': ' '.join(words[chunk_start : chunk_end + 1]),
            'type': chunk_type,
            'score': 1.0,
            'beginOffset': chunk_start,
            'endOffset': chunk_end,
        }
        res['tags'].append(tag)
    return res


def padding_sequence(seq, maxlen, padding = 'post', pad_int = 0):
    padded_seqs = []
    for s in seq:
        if padding == 'post':
            padded_seqs.append(s + [pad_int] * (maxlen - len(s)))
        if padding == 'pre':
            padded_seqs.append([pad_int] * (maxlen - len(s)) + s)
    return padded_seqs


def bert_tokenization(tokenizer, texts, cls = '[CLS]', sep = '[SEP]'):

    input_ids, input_masks, segment_ids, s_tokens = [], [], [], []
    for text in texts:
        text = remove_links_alias(text)
        tokens_a = tokenizer.tokenize(text)
        tokens_a = tokens_a if len(tokens_a) <= 510 else tokens_a[:510]
        tokens = [cls] + tokens_a + [sep]
        segment_id = [0] * len(tokens)
        input_id = tokenizer.convert_tokens_to_ids(tokens)
        input_mask = [1] * len(input_id)

        input_ids.append(input_id)
        input_masks.append(input_mask)
        segment_ids.append(segment_id)
        s_tokens.append(tokens)

    maxlen = max([len(i) for i in input_ids])
    input_ids = padding_sequence(input_ids, maxlen)
    input_masks = padding_sequence(input_masks, maxlen)
    segment_ids = padding_sequence(segment_ids, maxlen)

    return input_ids, input_masks, segment_ids, s_tokens


def _truncate_seq_pair(tokens_a, tokens_b, max_length):
    while True:
        total_length = len(tokens_a) + len(tokens_b)
        if total_length <= max_length:
            break
        if len(tokens_a) > len(tokens_b):
            tokens_a.pop()
        else:
            tokens_b.pop()


def bert_tokenization_siamese(
    tokenizer, left, right, cls = '[CLS]', sep = '[SEP]'
):
    input_ids, input_masks, segment_ids = [], [], []
    a, b = [], []
    for i in range(len(left)):
        tokens_a = tokenizer.tokenize(left[i])
        tokens_b = tokenizer.tokenize(right[i])
        a.append(tokens_a)
        b.append(tokens_b)

    maxlen = max([len(i) for i in a] + [len(i) for i in b]) + 5
    for i in range(len(left)):
        tokens_a = a[i]
        tokens_b = b[i]
        _truncate_seq_pair(tokens_a, tokens_b, maxlen - 3)

        tokens = []
        segment_id = []
        tokens.append(cls)
        segment_id.append(0)
        for token in tokens_a:
            tokens.append(token)
            segment_id.append(0)

        tokens.append(sep)
        segment_id.append(0)
        for token in tokens_b:
            tokens.append(token)
            segment_id.append(1)
        tokens.append(sep)
        segment_id.append(1)
        input_id = tokenizer.convert_tokens_to_ids(tokens)
        input_mask = [1] * len(input_id)

        while len(input_id) < maxlen:
            input_id.append(0)
            input_mask.append(0)
            segment_id.append(0)

        input_ids.append(input_id)
        input_masks.append(input_mask)
        segment_ids.append(segment_id)

    return input_ids, input_masks, segment_ids


SEG_ID_A = 0
SEG_ID_B = 1
SEG_ID_CLS = 2
SEG_ID_SEP = 3
SEG_ID_PAD = 4

special_symbols = {
    '<unk>': 0,
    '<s>': 1,
    '</s>': 2,
    '<cls>': 3,
    '<sep>': 4,
    '<pad>': 5,
    '<mask>': 6,
    '<eod>': 7,
    '<eop>': 8,
}

UNK_ID = special_symbols['<unk>']
CLS_ID = special_symbols['<cls>']
SEP_ID = special_symbols['<sep>']
MASK_ID = special_symbols['<mask>']
EOD_ID = special_symbols['<eod>']


def tokenize_fn(text, sp_model):
    text = preprocess_text(text, lower = False)
    return encode_ids(sp_model, text)


def xlnet_tokenization_siamese(tokenizer, left, right):
    input_ids, input_mask, all_seg_ids = [], [], []
    for i in range(len(left)):
        tokens = tokenize_fn(remove_links_alias(left[i]), tokenizer)
        tokens_right = tokenize_fn(remove_links_alias(right[i]), tokenizer)
        segment_ids = [SEG_ID_A] * len(tokens)
        tokens.append(SEP_ID)
        segment_ids.append(SEG_ID_A)

        tokens.extend(tokens_right)
        segment_ids.extend([SEG_ID_B] * len(tokens_right))
        tokens.append(SEP_ID)
        segment_ids.append(SEG_ID_B)

        tokens.append(CLS_ID)
        segment_ids.append(SEG_ID_CLS)

        cur_input_ids = tokens
        cur_input_mask = [0] * len(cur_input_ids)
        assert len(tokens) == len(cur_input_mask)
        assert len(tokens) == len(segment_ids)
        input_ids.append(tokens)
        input_mask.append(cur_input_mask)
        all_seg_ids.append(segment_ids)

    maxlen = max([len(i) for i in input_ids])
    input_ids = padding_sequence(input_ids, maxlen)
    input_mask = padding_sequence(input_mask, maxlen, pad_int = 1)
    all_seg_ids = padding_sequence(all_seg_ids, maxlen, pad_int = SEG_ID_PAD)
    return input_ids, input_mask, all_seg_ids


def xlnet_tokenization(tokenizer, texts):
    input_ids, input_masks, segment_ids, s_tokens = [], [], [], []
    for text in texts:
        text = remove_links_alias(text)
        tokens_a = tokenize_fn(text, tokenizer)
        tokens = []
        segment_id = []
        for token in tokens_a:
            tokens.append(token)
            segment_id.append(SEG_ID_A)

        tokens.append(SEP_ID)
        segment_id.append(SEG_ID_A)
        tokens.append(CLS_ID)
        segment_id.append(SEG_ID_CLS)

        input_id = tokens
        input_mask = [0] * len(input_id)

        input_ids.append(input_id)
        input_masks.append(input_mask)
        segment_ids.append(segment_id)
        s_tokens.append([tokenizer.IdToPiece(i) for i in tokens])

    maxlen = max([len(i) for i in input_ids])
    input_ids = padding_sequence(input_ids, maxlen)
    input_masks = padding_sequence(input_masks, maxlen, pad_int = 1)
    segment_ids = padding_sequence(segment_ids, maxlen, pad_int = SEG_ID_PAD)

    return input_ids, input_masks, segment_ids, s_tokens


def merge_wordpiece_tokens(paired_tokens, weighted = True):
    new_paired_tokens = []
    n_tokens = len(paired_tokens)

    i = 0

    while i < n_tokens:
        current_token, current_weight = paired_tokens[i]
        if current_token.startswith('##'):
            previous_token, previous_weight = new_paired_tokens.pop()
            merged_token = previous_token
            merged_weight = [previous_weight]
            while current_token.startswith('##'):
                merged_token = merged_token + current_token.replace('##', '')
                merged_weight.append(current_weight)
                i = i + 1
                current_token, current_weight = paired_tokens[i]
            merged_weight = np.mean(merged_weight)
            new_paired_tokens.append((merged_token, merged_weight))

        else:
            new_paired_tokens.append((current_token, current_weight))
            i = i + 1

    words = [
        i[0]
        for i in new_paired_tokens
        if i[0] not in ['[CLS]', '[SEP]', '[PAD]']
    ]
    weights = [
        i[1]
        for i in new_paired_tokens
        if i[0] not in ['[CLS]', '[SEP]', '[PAD]']
    ]
    if weighted:
        weights = np.array(weights)
        weights = weights / np.sum(weights)
    return list(zip(words, weights))


def merge_sentencepiece_tokens(paired_tokens, weighted = True):
    new_paired_tokens = []
    n_tokens = len(paired_tokens)
    rejected = ['<cls>', '<sep>']

    i = 0

    while i < n_tokens:

        current_token, current_weight = paired_tokens[i]
        if not current_token.startswith('▁') and current_token not in rejected:
            previous_token, previous_weight = new_paired_tokens.pop()
            merged_token = previous_token
            merged_weight = [previous_weight]
            while (
                not current_token.startswith('▁')
                and current_token not in rejected
            ):
                merged_token = merged_token + current_token.replace('▁', '')
                merged_weight.append(current_weight)
                i = i + 1
                current_token, current_weight = paired_tokens[i]
            merged_weight = np.mean(merged_weight)
            new_paired_tokens.append((merged_token, merged_weight))

        else:
            new_paired_tokens.append((current_token, current_weight))
            i = i + 1

    words = [
        i[0].replace('▁', '')
        for i in new_paired_tokens
        if i[0] not in ['<cls>', '<sep>', '<pad>']
    ]
    weights = [
        i[1]
        for i in new_paired_tokens
        if i[0] not in ['<cls>', '<sep>', '<pad>']
    ]
    if weighted:
        weights = np.array(weights)
        weights = weights / np.sum(weights)
    return list(zip(words, weights))


def parse_bert_tagging(left, tokenizer, cls = '[CLS]', sep = '[SEP]'):
    left = remove_links_alias(left)
    bert_tokens = [cls] + tokenizer.tokenize(left) + [sep]
    return tokenizer.convert_tokens_to_ids(bert_tokens), bert_tokens


def merge_wordpiece_tokens_tagging(x, y):
    new_paired_tokens = []
    n_tokens = len(x)

    i = 0
    while i < n_tokens:
        current_token, current_label = x[i], y[i]
        if current_token.startswith('##'):
            previous_token, previous_label = new_paired_tokens.pop()
            merged_token = previous_token
            merged_label = [previous_label]
            while current_token.startswith('##'):
                merged_token = merged_token + current_token.replace('##', '')
                merged_label.append(current_label)
                i = i + 1
                current_token, current_label = x[i], y[i]
            merged_label = merged_label[0]
            new_paired_tokens.append((merged_token, merged_label))
        else:
            new_paired_tokens.append((current_token, current_label))
            i = i + 1
    words = [
        i[0]
        for i in new_paired_tokens
        if i[0] not in ['[CLS]', '[SEP]', '[PAD]']
    ]
    labels = [
        i[1]
        for i in new_paired_tokens
        if i[0] not in ['[CLS]', '[SEP]', '[PAD]']
    ]
    return words, labels


def merge_sentencepiece_tokens_tagging(x, y):
    new_paired_tokens = []
    n_tokens = len(x)
    rejected = ['<cls>', '<sep>']

    i = 0

    while i < n_tokens:

        current_token, current_label = x[i], y[i]
        if not current_token.startswith('▁') and current_token not in rejected:
            previous_token, previous_label = new_paired_tokens.pop()
            merged_token = previous_token
            merged_label = [previous_label]
            while (
                not current_token.startswith('▁')
                and current_token not in rejected
            ):
                merged_token = merged_token + current_token.replace('▁', '')
                merged_label.append(current_label)
                i = i + 1
                current_token, current_label = x[i], y[i]
            merged_label = merged_label[0]
            new_paired_tokens.append((merged_token, merged_label))

        else:
            new_paired_tokens.append((current_token, current_label))
            i = i + 1

    words = [
        i[0].replace('▁', '')
        for i in new_paired_tokens
        if i[0] not in ['<cls>', '<sep>']
    ]
    labels = [i[1] for i in new_paired_tokens if i[0] not in ['<cls>', '<sep>']]
    return words, labels



import tensorflow as tf
import numpy as np


class LDA2VEC:
    def __init__(
        self,
        num_unique_documents,
        vocab_size,
        num_topics,
        freqs,
        embedding_size = 128,
        num_sampled = 40,
        learning_rate = 1e-3,
        lmbda = 150.0,
        alpha = None,
        power = 0.75,
        batch_size = 32,
        clip_gradients = 5.0,
        **kwargs
    ):
        moving_avgs = tf.train.ExponentialMovingAverage(0.9)
        self.batch_size = batch_size
        self.freqs = freqs
        self.sess = tf.InteractiveSession()

        self.X = tf.placeholder(tf.int32, shape = [None])
        self.Y = tf.placeholder(tf.int64, shape = [None])
        self.DOC = tf.placeholder(tf.int32, shape = [None])
        step = tf.Variable(0, trainable = False, name = 'global_step')
        self.switch_loss = tf.Variable(0, trainable = False)
        train_labels = tf.reshape(self.Y, [-1, 1])
        sampler = tf.nn.fixed_unigram_candidate_sampler(
            train_labels,
            num_true = 1,
            num_sampled = num_sampled,
            unique = True,
            range_max = vocab_size,
            distortion = power,
            unigrams = self.freqs,
        )

        self.word_embedding = tf.Variable(
            tf.random_uniform([vocab_size, embedding_size], -1.0, 1.0)
        )
        self.nce_weights = tf.Variable(
            tf.truncated_normal(
                [vocab_size, embedding_size],
                stddev = tf.sqrt(1 / embedding_size),
            )
        )
        self.nce_biases = tf.Variable(tf.zeros([vocab_size]))
        scalar = 1 / np.sqrt(num_unique_documents + num_topics)
        self.doc_embedding = tf.Variable(
            tf.random_normal(
                [num_unique_documents, num_topics],
                mean = 0,
                stddev = 50 * scalar,
            )
        )
        self.topic_embedding = tf.get_variable(
            'topic_embedding',
            shape = [num_topics, embedding_size],
            dtype = tf.float32,
            initializer = tf.orthogonal_initializer(gain = scalar),
        )
        pivot = tf.nn.embedding_lookup(self.word_embedding, self.X)
        proportions = tf.nn.embedding_lookup(self.doc_embedding, self.DOC)
        doc = tf.matmul(proportions, self.topic_embedding)
        doc_context = doc
        word_context = pivot
        context = tf.add(word_context, doc_context)
        loss_word2vec = tf.reduce_mean(
            tf.nn.nce_loss(
                weights = self.nce_weights,
                biases = self.nce_biases,
                labels = self.Y,
                inputs = context,
                num_sampled = num_sampled,
                num_classes = vocab_size,
                num_true = 1,
                sampled_values = sampler,
            )
        )
        self.fraction = tf.Variable(1, trainable = False, dtype = tf.float32)

        n_topics = self.doc_embedding.get_shape()[1].value
        log_proportions = tf.nn.log_softmax(self.doc_embedding)
        if alpha is None:
            alpha = 1.0 / n_topics
        loss = -(alpha - 1) * log_proportions
        prior = tf.reduce_sum(loss)

        loss_lda = lmbda * self.fraction * prior
        self.cost = tf.cond(
            step < self.switch_loss,
            lambda: loss_word2vec,
            lambda: loss_word2vec + loss_lda,
        )
        loss_avgs_op = moving_avgs.apply([loss_lda, loss_word2vec, self.cost])
        with tf.control_dependencies([loss_avgs_op]):
            self.optimizer = tf.contrib.layers.optimize_loss(
                self.cost,
                tf.train.get_global_step(),
                learning_rate,
                'Adam',
                clip_gradients = clip_gradients,
            )
        self.sess.run(tf.global_variables_initializer())

    def train(
        self, pivot_words, target_words, doc_ids, num_epochs, switch_loss = 3
    ):
        from tqdm import tqdm

        temp_fraction = self.batch_size / len(pivot_words)
        self.sess.run(tf.assign(self.fraction, temp_fraction))
        self.sess.run(tf.assign(self.switch_loss, switch_loss))
        for e in range(num_epochs):
            pbar = tqdm(
                range(0, len(pivot_words), self.batch_size),
                desc = 'minibatch loop',
            )
            for i in pbar:
                batch_x = pivot_words[
                    i : min(i + self.batch_size, len(pivot_words))
                ]
                batch_y = target_words[
                    i : min(i + self.batch_size, len(pivot_words))
                ]
                batch_doc = doc_ids[
                    i : min(i + self.batch_size, len(pivot_words))
                ]
                _, cost = self.sess.run(
                    [self.optimizer, self.cost],
                    feed_dict = {
                        self.X: batch_x,
                        self.Y: batch_y,
                        self.DOC: batch_doc,
                    },
                )
                pbar.set_postfix(cost = cost, epoch = e + 1)



from selenium import webdriver
import threading
from queue import Queue
from tqdm import tqdm
import warnings
import time

warnings.filterwarnings('ignore')

span = '/html/body/div[2]/div[1]/div[2]/div[1]/div[1]/div[2]/div[3]/div[1]/div[2]/div/span[1]'
base_url = 'https://translate.google.com/#'


class Translate:
    def __init__(self, from_lang, to_lang):
        self.driver = webdriver.PhantomJS()
        self.driver.set_page_load_timeout(30)
        final_url = base_url + from_lang.lower() + '/' + to_lang.lower()
        self.driver.get(final_url)

    def translate(self, string):
        assert isinstance(string, str), 'input must be a string'
        self.driver.find_element_by_id('source').clear()
        self.driver.find_element_by_id('source').send_keys(string)
        time.sleep(2)
        text = [elem.text for elem in self.driver.find_elements_by_xpath(span)]
        return text[0]


def task_translate(translator, string):
    return translator.translate(string)


def run_parallel_in_threads(args_list, target = task_translate):
    globalparas = []
    result = Queue()

    def task_wrapper(*args):
        result.put(target(*args))

    threads = [
        threading.Thread(target = task_wrapper, args = args)
        for args in args_list
    ]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    while not result.empty():
        globalparas.append(result.get())
    globalparas = list(filter(None, globalparas))
    return globalparas


class Translate_Concurrent:
    def __init__(self, batch_size, from_lang = 'en', to_lang = 'ms'):
        self._batch_size = batch_size
        self._translators = [
            Translate(from_lang, to_lang) for _ in range(self._batch_size)
        ]

    def translate_batch(self, strings):
        assert isinstance(strings, list) and isinstance(
            strings[0], str
        ), 'input must be list of strings'
        translated = []
        for no in tqdm(range(0, len(strings), self._batch_size), ncols = 70):
            data = strings[no : no + self._batch_size]
            combined = [
                (self._translators[i], data[i]) for i in range(len(data))
            ]
            translated.extend(run_parallel_in_threads(combined))
        return translated



import json
import os
from ._utils._utils import (
    check_file,
    load_graph,
    check_available,
    generate_session,
    sentencepiece_tokenizer_bert,
    sentencepiece_tokenizer_xlnet,
)
from . import home
from ._utils._paths import PATH_TOXIC, S3_PATH_TOXIC
from ._models._sklearn_model import MULTILABEL_BAYES
from ._models._bert_model import SIGMOID_BERT
from ._models._xlnet_model import SIGMOID_XLNET

from ._transformer._bert import bert_num_layers


_label_toxic = [
    'toxic',
    'severe_toxic',
    'obscene',
    'threat',
    'insult',
    'identity_hate',
]


_availability = {
    'bert': ['base', 'small'],
    'xlnet': ['base'],
    'albert': ['base'],
}


def available_transformer_model():
    """
    List available transformer toxicity analysis models.
    """
    return _availability


def multinomial(validate = True):
    """
    Load multinomial toxicity model.

    Parameters
    ----------
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    BAYES : malaya._models._sklearn_model.MULTILABEL_BAYES class
    """
    import pickle

    if validate:
        check_file(PATH_TOXIC['multinomial'], S3_PATH_TOXIC['multinomial'])
    else:
        if not check_available(PATH_TOXIC['multinomial']):
            raise Exception(
                'toxic/multinomial is not available, please `validate = True`'
            )
    try:
        with open(PATH_TOXIC['multinomial']['model'], 'rb') as fopen:
            multinomial = pickle.load(fopen)
        with open(PATH_TOXIC['multinomial']['vector'], 'rb') as fopen:
            vectorize = pickle.load(fopen)
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('toxic/multinomial') and try again"
        )
    from .stem import _classification_textcleaning_stemmer

    return MULTILABEL_BAYES(
        models = multinomial,
        vectors = vectorize,
        cleaning = _classification_textcleaning_stemmer,
    )


def transformer(model = 'xlnet', size = 'base', validate = True):
    """
    Load Transformer emotion model.

    Parameters
    ----------
    model : str, optional (default='bert')
        Model architecture supported. Allowed values:

        * ``'bert'`` - BERT architecture from google.
        * ``'xlnet'`` - XLNET architecture from google.
        * ``'albert'`` - ALBERT architecture from google.
    size : str, optional (default='base')
        Model size supported. Allowed values:

        * ``'base'`` - BASE size.
        * ``'small'`` - SMALL size.
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    BERT : malaya._models._bert_model.BINARY_BERT class
    """
    if not isinstance(model, str):
        raise ValueError('model must be a string')
    if not isinstance(size, str):
        raise ValueError('size must be a string')
    if not isinstance(validate, bool):
        raise ValueError('validate must be a boolean')

    model = model.lower()
    size = size.lower()
    if model not in _availability:
        raise Exception(
            'model not supported, please check supported models from malaya.sentiment.available_transformer_model()'
        )
    if size not in _availability[model]:
        raise Exception(
            'size not supported, please check supported models from malaya.sentiment.available_transformer_model()'
        )

    if validate:
        check_file(PATH_TOXIC[model][size], S3_PATH_TOXIC[model][size])
    else:
        if not check_available(PATH_TOXIC[model][size]):
            raise Exception(
                'toxicity/%s/%s is not available, please `validate = True`'
                % (model, size)
            )

    try:
        g = load_graph(PATH_TOXIC[model][size]['model'])
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('toxicity/%s/%s') and try again"
            % (model, size)
        )

    if model in ['albert', 'bert']:
        if model == 'bert':
            from ._transformer._bert import _extract_attention_weights_import
        if model == 'albert':
            from ._transformer._albert import _extract_attention_weights_import

        tokenizer, cls, sep = sentencepiece_tokenizer_bert(
            PATH_TOXIC[model][size]['tokenizer'],
            PATH_TOXIC[model][size]['vocab'],
        )

        return SIGMOID_BERT(
            X = g.get_tensor_by_name('import/Placeholder:0'),
            segment_ids = None,
            input_masks = None,
            logits = g.get_tensor_by_name('import/logits:0'),
            logits_seq = g.get_tensor_by_name('import/logits_seq:0'),
            sess = generate_session(graph = g),
            tokenizer = tokenizer,
            label = _label_toxic,
            cls = cls,
            sep = sep,
            attns = _extract_attention_weights_import(bert_num_layers[size], g),
            class_name = 'toxic',
        )
    if model in ['xlnet']:
        from ._transformer._xlnet import _extract_attention_weights_import

        tokenizer = sentencepiece_tokenizer_xlnet(
            PATH_TOXIC[model][size]['tokenizer']
        )

        return SIGMOID_XLNET(
            X = g.get_tensor_by_name('import/Placeholder:0'),
            segment_ids = g.get_tensor_by_name('import/Placeholder_1:0'),
            input_masks = g.get_tensor_by_name('import/Placeholder_2:0'),
            logits = g.get_tensor_by_name('import/logits:0'),
            logits_seq = g.get_tensor_by_name('import/logits_seq:0'),
            sess = generate_session(graph = g),
            tokenizer = tokenizer,
            label = _label_toxic,
            attns = _extract_attention_weights_import(g),
            class_name = 'toxic',
        )



import setuptools


__packagename__ = 'malaya-gpu'

setuptools.setup(
    name = __packagename__,
    packages = setuptools.find_packages(),
    version = '3.1.2',
    python_requires = '>=3.6.*',
    description = 'Natural-Language-Toolkit for bahasa Malaysia, powered by Deep Learning Tensorflow. GPU Version',
    author = 'huseinzol05',
    author_email = 'husein.zol05@gmail.com',
    url = 'https://github.com/huseinzol05/Malaya',
    download_url = 'https://github.com/huseinzol05/Malaya/archive/master.zip',
    keywords = ['nlp', 'bm'],
    install_requires = [
        'dateparser',
        'sklearn',
        'scikit-learn',
        'requests',
        'unidecode',
        'tensorflow-gpu==1.15',
        'numpy',
        'scipy',
        'PySastrawi',
        'ftfy',
        'networkx',
        'sentencepiece',
        'bert-tensorflow',
        'tqdm',
    ],
    license = 'MIT',
    classifiers = [
        'Programming Language :: Python :: 3.6',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Text Processing',
    ],
    package_data = {
        'malaya': [
            '_utils/web/*.html',
            '_utils/web/static/*.js',
            '_utils/web/static/*.css',
        ]
    },
)



import re
from ..num2word import to_cardinal, to_ordinal
from ..word2num import word2num
from ._tatabahasa import (
    rules_normalizer,
    consonants,
    vowels,
    sounds,
    hujung_malaysian,
    calon_dictionary,
)

ignore_words = ['ringgit', 'sen']
ignore_postfix = ['adalah']


def _remove_postfix(word):
    if word in ignore_postfix:
        return word, ''
    for p in hujung_malaysian:
        if word.endswith(p):
            return word[: -len(p)], ' lah'
    return word, ''


def _normalize_title(word):
    if word[0].isupper():
        return calon_dictionary.get(word.lower(), word)
    return word


def _normalize_money(
    word, currency = {'dollar': '$', 'ringgit': 'RM', 'pound': '£', 'euro': '€'}
):
    splitted = word.split()
    if splitted[-1] in ['dollar', 'ringgit', 'pound', 'euro']:
        v = word2num(' '.join(splitted[:-1]))
        return currency[splitted[-1]] + str(v)
    else:
        return word


def _is_number_regex(s):
    if re.match('^\d+?\.\d+?$', s) is None:
        return s.isdigit()
    return True


def _string_to_num(word):
    if '.' in word:
        return float(word)
    else:
        return int(word)


def cardinal(x):
    try:
        if re.match('.*[A-Za-z]+.*', x):
            return x
        x = re.sub(',', '', x, count = 10)

        if re.match('.+\..*', x):
            x = to_cardinal(float(x))
        elif re.match('\..*', x):
            x = to_cardinal(float(x))
        else:
            x = to_cardinal(int(x))
        x = re.sub('-', ' ', x, count = 10)
        return x
    except:
        return x


def digit(x):
    try:
        x = re.sub('[^0-9]', '', x)
        result_string = ''
        for i in x:
            result_string = result_string + cardinal(i) + ' '
        result_string = result_string.strip()
        return result_string
    except:
        return x


def letters(x):
    try:
        x = re.sub('[^a-zA-Z]', '', x)
        x = x.lower()
        result_string = ''
        for i in range(len(x)):
            result_string = result_string + x[i] + ' '
        return result_string.strip()
    except:
        return x


def rom_to_int(string):

    table = [
        ['M', 1000],
        ['CM', 900],
        ['D', 500],
        ['CD', 400],
        ['C', 100],
        ['XC', 90],
        ['L', 50],
        ['XL', 40],
        ['X', 10],
        ['IX', 9],
        ['V', 5],
        ['IV', 4],
        ['I', 1],
    ]
    returnint = 0
    for pair in table:

        continueyes = True

        while continueyes:
            if len(string) >= len(pair[0]):

                if string[0 : len(pair[0])] == pair[0]:
                    returnint += pair[1]
                    string = string[len(pair[0]) :]

                else:
                    continueyes = False
            else:
                continueyes = False

    return returnint


def ordinal(x):
    try:
        result_string = ''
        x = x.replace(',', '')
        x = x.replace('[\.]$', '')
        if re.match('^[0-9]+$', x):
            x = to_ordinal(int(x))
            return x
        if re.match('.*(V|X|I|L|D)', x):
            x = x.replace('-', '')
            if re.match('^ke.*', x):
                x = x[2:]
                x = rom_to_int(x)
                result_string = to_ordinal(x)
            else:
                x = rom_to_int(x)
                result_string = to_ordinal(x)
                result_string = 'yang ' + result_string
        elif re.match('^ke.*', x):
            x = x.replace('-', '')
            x = x[2:]
            result_string = to_ordinal(int(x))
        else:
            result_string = to_ordinal(int(x))
        return result_string
    except Exception as e:
        return x


def telephone(x):
    try:
        result_string = ''
        for i in range(0, len(x)):
            if re.match('[0-9]+', x[i]):
                result_string = result_string + cardinal(x[i]) + ' '
            else:
                result_string = result_string + 'sil '
        return result_string.strip()
    except:
        return x


def electronic(x):
    try:
        replacement = {
            '.': 'dot',
            ':': 'colon',
            '/': 'slash',
            '-': 'dash',
            '#': 'hash tag',
        }
        result_string = ''
        if re.match('.*[A-Za-z].*', x):
            for char in x:
                if re.match('[A-Za-z]', char):
                    result_string = result_string + letters(char) + ' '
                elif char in replacement:
                    result_string = result_string + replacement[char] + ' '
                elif re.match('[0-9]', char):
                    if char == 0:
                        result_string = result_string + 'o '
                    else:
                        number = cardinal(char)
                        for n in number:
                            result_string = result_string + n + ' '
            return result_string.strip()
        else:
            return x
    except:
        return x


def fraction(x):
    try:
        y = x.split('/')
        result_string = ''
        y[0] = cardinal(y[0])
        y[1] = cardinal(y[1])
        return '%s per %s' % (y[0], y[1])
    except:
        return x


def money(x):
    try:
        if (
            re.match('^\$', x)
            or x.lower().endswith('dollar')
            or x.lower().endswith('cent')
        ):
            x = x.lower()
            if not x.startswith('$') and x.endswith('cent'):
                cent = True
            else:
                cent = False
            x = x.replace('$', '').replace('dollar', '').replace('cent', '')
            x = re.sub(r'[ ]+', ' ', x).strip()
            x, n = re.split("(\d+(?:[\.,']\d+)?)", x)[1:]
            x = re.sub(',', '', x, count = 10)
            labels = []
            for c in n:
                if re.match('.*(M|m)$', c):
                    labels.append(1e6)
                elif re.match('.*(b|B)$', c):
                    labels.append(1e9)
                elif re.match('.*(k|K)$', c):
                    labels.append(1e3)

            x = float(x)
            if cent:
                x = x / 100
            for l in labels:
                x = x * l

            text = cardinal(str(x))
            c = '$%s' % (str(x))
            x = '%s dollar' % (text)
            return re.sub(r'[ ]+', ' ', x.lower()).strip(), c

        elif (
            re.match('^US', x)
            or x.lower().endswith('dollar')
            or x.lower().endswith('cent')
        ):
            x = x.lower()
            if not x.startswith('US') and x.endswith('cent'):
                cent = True
            else:
                cent = False
            x = x.replace('US', '').replace('dollar', '').replace('cent', '')
            x = re.sub(r'[ ]+', ' ', x).strip()
            x, n = re.split("(\d+(?:[\.,']\d+)?)", x)[1:]
            x = re.sub(',', '', x, count = 10)
            labels = []
            for c in n:
                if re.match('.*(M|m)$', c):
                    labels.append(1e6)
                elif re.match('.*(b|B)$', c):
                    labels.append(1e9)
                elif re.match('.*(k|K)$', c):
                    labels.append(1e3)

            x = float(x)
            if cent:
                x = x / 100
            for l in labels:
                x = x * l

            text = cardinal(str(x))
            c = '$%s' % (str(x))
            x = '%s dollar' % (text)
            return re.sub(r'[ ]+', ' ', x.lower()).strip(), c

        elif (
            re.match('^\£', x)
            or x.lower().endswith('pound')
            or x.lower().endswith('penny')
        ):
            x = x.lower()
            if not x.startswith('£') and x.endswith('penny'):
                cent = True
            else:
                cent = False
            x = x.replace('£', '').replace('pound', '').replace('penny', '')
            x = re.sub(r'[ ]+', ' ', x).strip()
            x, n = re.split("(\d+(?:[\.,']\d+)?)", x)[1:]
            x = re.sub(',', '', x, count = 10)
            labels = []
            for c in n:
                if re.match('.*(M|m)$', c):
                    labels.append(1e6)
                elif re.match('.*(b|B)$', c):
                    labels.append(1e9)
                elif re.match('.*(k|K)$', c):
                    labels.append(1e3)

            x = float(x)
            if cent:
                x = x / 100
            for l in labels:
                x = x * l

            text = cardinal(str(x))
            c = '£%s' % (str(x))
            x = '%s pound' % (text)
            return re.sub(r'[ ]+', ' ', x.lower()).strip(), c

        elif (
            re.match('^\€', x)
            or x.lower().endswith('euro')
            or x.lower().endswith('cent')
        ):
            x = x.lower()
            if not x.startswith('€') and x.endswith('cent'):
                cent = True
            else:
                cent = False
            x = x.replace('€', '').replace('euro', '').replace('cent', '')
            x = re.sub(r'[ ]+', ' ', x).strip()
            x, n = re.split("(\d+(?:[\.,']\d+)?)", x)[1:]
            x = re.sub(',', '', x, count = 10)
            labels = []
            for c in n:
                if re.match('.*(M|m)$', c):
                    labels.append(1e6)
                elif re.match('.*(b|B)$', c):
                    labels.append(1e9)
                elif re.match('.*(k|K)$', c):
                    labels.append(1e3)

            x = float(x)
            if cent:
                x = x / 100
            for l in labels:
                x = x * l

            text = cardinal(str(x))
            c = '€%s' % (str(x))
            x = '%s euro' % (text)
            return re.sub(r'[ ]+', ' ', x.lower()).strip(), c

        elif (
            re.match('^RM', x)
            or re.match('^rm', x)
            or x.lower().endswith('ringgit')
            or x.lower().endswith('sen')
        ):
            x = x.lower()
            if not x.startswith('rm') and x.endswith('sen'):
                cent = True
            else:
                cent = False
            x = x.replace('rm', '').replace('ringgit', '').replace('sen', '')
            x = re.sub(r'[ ]+', ' ', x).strip()
            x, n = re.split("(\d+(?:[\.,']\d+)?)", x)[1:]
            x = re.sub(',', '', x, count = 10)
            labels = []
            for c in n:
                if re.match('.*(M|m)$', c):
                    labels.append(1e6)
                elif re.match('.*(b|B)$', c):
                    labels.append(1e9)
                elif re.match('.*(k|K)$', c):
                    labels.append(1e3)

            x = float(x)
            if cent:
                x = x / 100
            for l in labels:
                x = x * l

            text = cardinal(str(x))
            c = 'RM%s' % (str(x))
            x = '%s ringgit' % (text)
            return re.sub(r'[ ]+', ' ', x.lower()).strip(), c
        return x, None

    except Exception as e:
        return x, None



import tensorflow as tf
from ._xlnet_model import xlnet as xlnet_lib
from ..texts._text_functions import (
    xlnet_tokenization,
    padding_sequence,
    merge_sentencepiece_tokens,
)
import collections
from collections import defaultdict
import re
import os
import numpy as np


def get_assignment_map_from_checkpoint(tvars, init_checkpoint):
    assignment_map = {}
    initialized_variable_names = {}

    name_to_variable = collections.OrderedDict()
    for var in tvars:
        name = var.name
        m = re.match('^(.*):\\d+$', name)
        if m is not None:
            name = m.group(1)
        name_to_variable[name] = var

    init_vars = tf.train.list_variables(init_checkpoint)

    assignment_map = collections.OrderedDict()
    for x in init_vars:
        (name, var) = (x[0], x[1])
        if name not in name_to_variable:
            continue
        assignment_map[name] = name_to_variable[name]
        initialized_variable_names[name] = 1
        initialized_variable_names[name + ':0'] = 1

    return (assignment_map, initialized_variable_names)


def _extract_attention_weights_import(tf_graph):
    attentions = [
        n.name
        for n in tf_graph.as_graph_def().node
        if 'rel_attn/Softmax' in n.name
    ]

    return [tf_graph.get_tensor_by_name('%s:0' % (a)) for a in attentions]


class _Model:
    def __init__(self, xlnet_config, tokenizer, checkpoint, pool_mode = 'last'):

        kwargs = dict(
            is_training = True,
            use_tpu = False,
            use_bfloat16 = False,
            dropout = 0.0,
            dropatt = 0.0,
            init = 'normal',
            init_range = 0.1,
            init_std = 0.05,
            clamp_len = -1,
        )

        xlnet_parameters = xlnet_lib.RunConfig(**kwargs)

        self._tokenizer = tokenizer
        _graph = tf.Graph()
        with _graph.as_default():
            self.X = tf.placeholder(tf.int32, [None, None])
            self.segment_ids = tf.placeholder(tf.int32, [None, None])
            self.input_masks = tf.placeholder(tf.float32, [None, None])

            xlnet_model = xlnet_lib.XLNetModel(
                xlnet_config = xlnet_config,
                run_config = xlnet_parameters,
                input_ids = tf.transpose(self.X, [1, 0]),
                seg_ids = tf.transpose(self.segment_ids, [1, 0]),
                input_mask = tf.transpose(self.input_masks, [1, 0]),
            )

            self.logits = xlnet_model.get_pooled_out(pool_mode, True)
            self._sess = tf.InteractiveSession()
            self._sess.run(tf.global_variables_initializer())
            tvars = tf.trainable_variables()
            assignment_map, _ = get_assignment_map_from_checkpoint(
                tvars, checkpoint
            )
            self._saver = tf.train.Saver(var_list = assignment_map)
            attentions = [
                n.name
                for n in tf.get_default_graph().as_graph_def().node
                if 'rel_attn/Softmax' in n.name
            ]
            g = tf.get_default_graph()
            self.attention_nodes = [
                g.get_tensor_by_name('%s:0' % (a)) for a in attentions
            ]

    def vectorize(self, strings):
        """
        Vectorize string inputs using bert attention.

        Parameters
        ----------
        strings : str / list of str

        Returns
        -------
        array: vectorized strings
        """

        if isinstance(strings, list):
            if not isinstance(strings[0], str):
                raise ValueError('input must be a list of strings or a string')
        else:
            if not isinstance(strings, str):
                raise ValueError('input must be a list of strings or a string')
        if isinstance(strings, str):
            strings = [strings]

        input_ids, input_masks, segment_ids, _ = xlnet_tokenization(
            self._tokenizer, strings
        )
        return self._sess.run(
            self.logits,
            feed_dict = {
                self.X: input_ids,
                self.segment_ids: segment_ids,
                self.input_masks: input_masks,
            },
        )

    def _attention(self, strings):
        input_ids, input_masks, segment_ids, s_tokens = xlnet_tokenization(
            self._tokenizer, strings
        )
        maxlen = max([len(s) for s in s_tokens])
        s_tokens = padding_sequence(s_tokens, maxlen, pad_int = '<cls>')
        attentions = self._sess.run(
            self.attention_nodes,
            feed_dict = {
                self.X: input_ids,
                self.segment_ids: segment_ids,
                self.input_masks: input_masks,
            },
        )
        return attentions, s_tokens

    def attention(self, strings, method = 'last', **kwargs):
        """
        Get attention string inputs from xlnet attention.

        Parameters
        ----------
        strings : str / list of str
        method : str, optional (default='last')
            Attention layer supported. Allowed values:

            * ``'last'`` - attention from last layer.
            * ``'first'`` - attention from first layer.
            * ``'mean'`` - average attentions from all layers.

        Returns
        -------
        array: attention
        """

        if isinstance(strings, list):
            if not isinstance(strings[0], str):
                raise ValueError('input must be a list of strings or a string')
        else:
            if not isinstance(strings, str):
                raise ValueError('input must be a list of strings or a string')
        if isinstance(strings, str):
            strings = [strings]

        method = method.lower()
        if method not in ['last', 'first', 'mean']:
            raise Exception(
                "method not supported, only support ['last', 'first', 'mean']"
            )
        attentions, s_tokens = self._attention(strings)

        if method == 'first':
            cls_attn = np.transpose(attentions[0][:, 0], (1, 0, 2))

        if method == 'last':
            cls_attn = np.transpose(attentions[-1][:, 0], (1, 0, 2))

        if method == 'mean':
            cls_attn = np.transpose(
                np.mean(attentions, axis = 0).mean(axis = 1), (1, 0, 2)
            )

        cls_attn = np.mean(cls_attn, axis = 1)
        total_weights = np.sum(cls_attn, axis = -1, keepdims = True)
        attn = cls_attn / total_weights
        output = []
        for i in range(attn.shape[0]):
            output.append(
                merge_sentencepiece_tokens(list(zip(s_tokens[i], attn[i])))
            )
        return output

    def visualize_attention(self, string):
        from .._utils._html import _attention

        if not isinstance(string, str):
            raise ValueError('input must be a string')
        strings = [string]
        attentions, s_tokens = self._attention(strings)
        attn_dict = defaultdict(list)
        for layer, attn_data in enumerate(attentions):
            attn = attn_data[:, :, 0]
            attn_dict['all'].append(attn.tolist())

        results = {
            'all': {
                'attn': attn_dict['all'],
                'left_text': s_tokens[0],
                'right_text': s_tokens[0],
            }
        }
        _attention(results)


def xlnet(model = 'base', pool_mode = 'last', validate = True):
    """
    Load xlnet model.

    Parameters
    ----------
    model : str, optional (default='base')
        Model architecture supported. Allowed values:

        * ``'base'`` - base xlnet-bahasa released by Malaya.
    pool_mode : str, optional (default='last')
        Model logits architecture supported. Allowed values:

        * ``'last'`` - last of the sequence.
        * ``'first'`` - first of the sequence.
        * ``'mean'`` - mean of the sequence.
        * ``'attn'`` - attention of the sequence.
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    XLNET_MODEL: malaya._transformer._xlnet._Model class
    """

    if not isinstance(model, str):
        raise ValueError('model must be a string')
    if not isinstance(pool_mode, str):
        raise ValueError('pool_mode must be a string')
    if not isinstance(validate, bool):
        raise ValueError('validate must be a boolean')

    model = model.lower()
    pool_mode = pool_mode.lower()

    from .._utils._paths import PATH_XLNET, S3_PATH_XLNET
    from .._utils._utils import check_file, check_available

    if pool_mode not in ['last', 'first', 'mean', 'attn']:
        raise Exception(
            "pool_mode not supported, only support ['last', 'first', 'mean', 'attn']"
        )

    if validate:
        check_file(PATH_XLNET[model]['model'], S3_PATH_XLNET[model])
    else:
        if not check_available(PATH_XLNET[model]['model']):
            raise Exception(
                'xlnet-model/%s is not available, please `validate = True`'
                % (model)
            )

    if not os.path.exists(PATH_XLNET[model]['directory']):
        import tarfile

        with tarfile.open(PATH_XLNET[model]['model']['model']) as tar:
            tar.extractall(path = PATH_XLNET[model]['path'])

    import sentencepiece as spm

    sp_model = spm.SentencePieceProcessor()
    sp_model.Load(PATH_XLNET[model]['directory'] + 'sp10m.cased.v5.model')
    xlnet_config = xlnet_lib.XLNetConfig(
        json_path = PATH_XLNET[model]['directory'] + 'config.json'
    )
    xlnet_checkpoint = PATH_XLNET[model]['directory'] + 'model.ckpt'
    model = _Model(
        xlnet_config, sp_model, xlnet_checkpoint, pool_mode = pool_mode
    )
    model._saver.restore(model._sess, xlnet_checkpoint)
    return model



import numpy as np
import re
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from sklearn.decomposition import TruncatedSVD, NMF, LatentDirichletAllocation
from .texts._jarowrinkler import JaroWinkler
from sklearn.metrics.pairwise import cosine_similarity
from .texts._text_functions import (
    summary_textcleaning,
    STOPWORDS,
    split_into_sentences,
    simple_textcleaning,
)
import itertools
import networkx as nx
from .stem import sastrawi
from ._models import _skip_thought
from .cluster import cluster_words
from .texts.vectorizer import SkipGramVectorizer


class _DEEP_SKIPTHOUGHT:
    def __init__(
        self, sess, x, logits, attention, dictionary, maxlen, model = None
    ):
        self._sess = sess
        self._X = x
        self._logits = logits
        self._attention = attention
        self.dictionary = dictionary
        self._maxlen = maxlen
        self._rev_dictionary = {v: k for k, v in self.dictionary.items()}
        self._model = model

    def vectorize(self, strings):

        """
        Vectorize string inputs using bert attention.

        Parameters
        ----------
        strings : str / list of str

        Returns
        -------
        array: vectorized strings
        """

        if isinstance(strings, list):
            if not isinstance(strings[0], str):
                raise ValueError('input must be a list of strings or a string')
        else:
            if not isinstance(strings, str):
                raise ValueError('input must be a list of strings or a string')
        if isinstance(strings, str):
            strings = [strings]

        splitted_fullstop = [summary_textcleaning(i) for i in strings]
        original_strings = [i[0] for i in splitted_fullstop]
        cleaned_strings = [i[1] for i in splitted_fullstop]
        sequences = _skip_thought.batch_sequence(
            cleaned_strings, self.dictionary, maxlen = self._maxlen
        )
        return self._sess.run(
            self._logits, feed_dict = {self._X: np.array(sequences)}
        )


class _DEEP_SUMMARIZER:
    def __init__(self, vectorizer):
        self._vectorizer = vectorizer

    def summarize(self, corpus, top_k = 3, important_words = 3, **kwargs):
        """
        Summarize list of strings / corpus

        Parameters
        ----------
        corpus: str, list

        top_k: int, (default=3)
            number of summarized strings.
        important_words: int, (default=3)
            number of important words.

        Returns
        -------
        string: summarized string
        """
        if not isinstance(top_k, int):
            raise ValueError('top_k must be an integer')
        if not isinstance(important_words, int):
            raise ValueError('important_words must be an integer')
        if not isinstance(corpus, list) and not isinstance(corpus, str):
            raise ValueError('corpus must be a list')
        if isinstance(corpus, list):
            if not isinstance(corpus[0], str):
                raise ValueError('corpus must be list of strings')

        if isinstance(corpus, str):
            corpus = split_into_sentences(corpus)
        else:
            corpus = '. '.join(corpus)
            corpus = split_into_sentences(corpus)

        splitted_fullstop = [summary_textcleaning(i) for i in corpus]
        original_strings = [i[0] for i in splitted_fullstop]
        cleaned_strings = [i[1] for i in splitted_fullstop]

        if '_DEEP_SKIPTHOUGHT' in str(self._vectorizer):

            sequences = _skip_thought.batch_sequence(
                cleaned_strings,
                self._vectorizer.dictionary,
                maxlen = self._vectorizer._maxlen,
            )
            vectors, attention = self._vectorizer._sess.run(
                [self._vectorizer._logits, self._vectorizer._attention],
                feed_dict = {self._vectorizer._X: np.array(sequences)},
            )
            attention = attention.sum(axis = 0)
            indices = np.argsort(attention)[::-1]
            top_words = [
                self._vectorizer._rev_dictionary[i]
                for i in indices
                if self._vectorizer._rev_dictionary[i] not in STOPWORDS
            ][:important_words]

        else:
            vectors = self._vectorizer.vectorize(corpus)
            attentions = self._vectorizer.attention(corpus, **kwargs)
            flatten = list(itertools.chain(*attentions))
            r = {}
            for f in flatten:
                c = simple_textcleaning(f[0])
                if c in STOPWORDS:
                    continue
                if c not in r:
                    r[c] = f[1]
                else:
                    r[c] += f[1]
            top_words = sorted(r, key = r.get, reverse = True)[:important_words]

        similar = cosine_similarity(vectors, vectors)
        similar[similar >= 0.99999] = 0
        nx_graph = nx.from_numpy_array(similar)
        scores = nx.pagerank(nx_graph, max_iter = 10000)
        ranked_sentences = sorted(
            ((scores[i], s) for i, s in enumerate(original_strings)),
            reverse = True,
        )
        summary = [r[1] for r in ranked_sentences[:top_k]]

        return {
            'summary': ' '.join(summary),
            'top-words': top_words,
            'cluster-top-words': cluster_words(top_words),
        }


def available_skipthought():
    """
    List available deep skip-thought models.
    """
    return ['lstm', 'residual-network']


def deep_skipthought(model = 'lstm'):
    """
    Load deep learning skipthought model.

    Parameters
    ----------
    model : str, optional (default='skip-thought')
        Model architecture supported. Allowed values:

        * ``'lstm'`` - LSTM skip-thought deep learning model trained on news dataset. Hopefully we can train on wikipedia dataset.
        * ``'residual-network'`` - residual network with Bahdanau Attention skip-thought deep learning model trained on wikipedia dataset.

    Returns
    -------
    _DEEP_SKIPTHOUGHT: malaya.summarize._DEEP_SKIPTHOUGHT class
    """
    model = model.lower()
    if model == 'lstm':
        model = _skip_thought.news_load_model
    elif model == 'residual-network':
        model = _skip_thought.wiki_load_model
    else:
        raise Exception(
            'model is not supported, please check supported models from malaya.summarize.available_skipthought()'
        )
    sess, x, logits, attention, dictionary, maxlen = model()
    return _DEEP_SKIPTHOUGHT(sess, x, logits, attention, dictionary, maxlen)


def encoder(vectorizer):
    """
    Encoder interface for text similarity.

    Parameters
    ----------
    vectorizer : object
        encoder interface object, BERT, skip-thought, XLNET.

    Returns
    -------
    _DOC2VEC_SIMILARITY: malaya.similarity._DOC2VEC_SIMILARITY
    """

    if not hasattr(vectorizer, 'vectorize'):
        raise ValueError('vectorizer must has `vectorize` method')
    return _DEEP_SUMMARIZER(vectorizer)


def _base_summarizer(
    corpus,
    decomposition,
    top_k = 3,
    max_df = 0.95,
    min_df = 2,
    ngram = (1, 3),
    vectorizer = 'bow',
    important_words = 10,
    **kwargs,
):
    if not isinstance(vectorizer, str):
        raise ValueError('vectorizer must be a string')
    if not isinstance(top_k, int):
        raise ValueError('top_k must be an integer')
    vectorizer = vectorizer.lower()
    if not vectorizer in ['tfidf', 'bow', 'skip-gram']:
        raise ValueError("vectorizer must be in  ['tfidf', 'bow', 'skip-gram']")
    if not isinstance(ngram, tuple):
        raise ValueError('ngram must be a tuple')
    if not len(ngram) == 2:
        raise ValueError('ngram size must equal to 2')
    if not isinstance(min_df, int):
        raise ValueError('min_df must be an integer')
    if not (isinstance(max_df, int) or isinstance(max_df, float)):
        raise ValueError('max_df must be an integer or a float')
    if min_df < 1:
        raise ValueError('min_df must be bigger than 0')
    if not (max_df <= 1 and max_df > 0):
        raise ValueError(
            'max_df must be bigger than 0, less than or equal to 1'
        )
    if not isinstance(corpus, list) and not isinstance(corpus, str):
        raise ValueError('corpus must be a list')
    if isinstance(corpus, list):
        if not isinstance(corpus[0], str):
            raise ValueError('corpus must be list of strings')
    if isinstance(corpus, str):
        corpus = split_into_sentences(corpus)
    else:
        corpus = '. '.join(corpus)
        corpus = split_into_sentences(corpus)

    splitted_fullstop = [summary_textcleaning(i) for i in corpus]
    original_strings = [i[0] for i in splitted_fullstop]
    cleaned_strings = [i[1] for i in splitted_fullstop]
    stemmed = [sastrawi(i) for i in cleaned_strings]

    if vectorizer == 'tfidf':
        Vectorizer = TfidfVectorizer
    elif vectorizer == 'bow':
        Vectorizer = CountVectorizer
    elif vectorizer == 'skip-gram':
        Vectorizer = SkipGramVectorizer
    else:
        raise Exception("vectorizer must be in  ['tfidf', 'bow', 'skip-gram']")
    tf_vectorizer = Vectorizer(
        max_df = max_df,
        min_df = min_df,
        ngram_range = ngram,
        stop_words = STOPWORDS,
        **kwargs,
    )
    tf = tf_vectorizer.fit_transform(stemmed)
    if hasattr(tf_vectorizer, 'idf_'):
        indices = np.argsort(tf_vectorizer.idf_)[::-1]
    else:
        indices = np.argsort(np.asarray(tf.sum(axis = 0))[0])[::-1]

    features = tf_vectorizer.get_feature_names()
    top_words = [features[i] for i in indices[:important_words]]
    vectors = decomposition(tf.shape[1] // 2).fit_transform(tf)
    similar = cosine_similarity(vectors, vectors)
    similar[similar >= 0.999] = 0
    nx_graph = nx.from_numpy_array(similar)
    scores = nx.pagerank(nx_graph, max_iter = 10000)
    ranked_sentences = sorted(
        ((scores[i], s) for i, s in enumerate(original_strings)), reverse = True
    )
    summary = [r[1] for r in ranked_sentences[:top_k]]
    return {
        'summary': ' '.join(summary),
        'top-words': top_words,
        'cluster-top-words': cluster_words(top_words),
    }


def lda(
    corpus,
    top_k = 3,
    important_words = 10,
    max_df = 0.95,
    min_df = 2,
    ngram = (1, 3),
    vectorizer = 'bow',
    **kwargs,
):
    """
    summarize a list of strings using LDA, scoring using TextRank.

    Parameters
    ----------
    corpus: list
    top_k: int, (default=3)
        number of summarized strings.
    important_words: int, (default=10)
        number of important words.
    max_df: float, (default=0.95)
        maximum of a word selected based on document frequency.
    min_df: int, (default=2)
        minimum of a word selected on based on document frequency.
    ngram: tuple, (default=(1,3))
        n-grams size to train a corpus.
    vectorizer: str, (default='bow')
        vectorizer technique. Allowed values:

        * ``'bow'`` - Bag of Word.
        * ``'tfidf'`` - Term frequency inverse Document Frequency.
        * ``'skip-gram'`` - Bag of Word with skipping certain n-grams.

    Returns
    -------
    dict: result
    """
    return _base_summarizer(
        corpus,
        LatentDirichletAllocation,
        top_k = top_k,
        max_df = max_df,
        min_df = min_df,
        ngram = ngram,
        vectorizer = vectorizer,
        important_words = important_words,
        **kwargs,
    )


def lsa(
    corpus,
    top_k = 3,
    important_words = 10,
    max_df = 0.95,
    min_df = 2,
    ngram = (1, 3),
    vectorizer = 'bow',
    **kwargs,
):
    """
    summarize a list of strings using LSA, scoring using TextRank.

    Parameters
    ----------
    corpus: list
    top_k: int, (default=3)
        number of summarized strings.
    important_words: int, (default=10)
        number of important words.
    max_df: float, (default=0.95)
        maximum of a word selected based on document frequency.
    min_df: int, (default=2)
        minimum of a word selected on based on document frequency.
    ngram: tuple, (default=(1,3))
        n-grams size to train a corpus.
    vectorizer: str, (default='bow')
        vectorizer technique. Allowed values:

        * ``'bow'`` - Bag of Word.
        * ``'tfidf'`` - Term frequency inverse Document Frequency.
        * ``'skip-gram'`` - Bag of Word with skipping certain n-grams.

    Returns
    -------
    dict: result
    """
    return _base_summarizer(
        corpus,
        TruncatedSVD,
        top_k = top_k,
        max_df = max_df,
        min_df = min_df,
        ngram = ngram,
        vectorizer = vectorizer,
        important_words = important_words,
        **kwargs,
    )


def doc2vec(vectorizer, corpus, top_k = 3, aggregation = 'mean', soft = True):
    """
    summarize a list of strings using doc2vec, scoring using TextRank.

    Parameters
    ----------
    vectorizer : object
        fast-text or word2vec interface object.
    corpus: list
    top_k: int, (default=3)
        number of summarized strings.
    aggregation : str, optional (default='mean')
        Aggregation supported. Allowed values:

        * ``'mean'`` - mean.
        * ``'min'`` - min.
        * ``'max'`` - max.
        * ``'sum'`` - sum.
        * ``'sqrt'`` - square root.
    soft: bool, optional (default=True)
        word not inside vectorizer will replace with nearest word if True, else, will skip.

    Returns
    -------
    dictionary: result
    """
    if not hasattr(vectorizer, 'get_vector_by_name'):
        raise ValueError('vectorizer must has `get_vector_by_name` method')
    if not isinstance(top_k, int):
        raise ValueError('top_k must be an integer')
    if not isinstance(corpus, list) and not isinstance(corpus, str):
        raise ValueError('corpus must be a list')
    if isinstance(corpus, list):
        if not isinstance(corpus[0], str):
            raise ValueError('corpus must be list of strings')
    if isinstance(corpus, str):
        corpus = split_into_sentences(corpus)
    else:
        corpus = '. '.join(corpus)
        corpus = split_into_sentences(corpus)
    splitted_fullstop = [summary_textcleaning(i) for i in corpus]
    original_strings = [i[0] for i in splitted_fullstop]
    cleaned_strings = [i[1] for i in splitted_fullstop]

    aggregation = aggregation.lower()
    if aggregation == 'mean':
        aggregation_function = np.mean
    elif aggregation == 'min':
        aggregation_function = np.min
    elif aggregation == 'max':
        aggregation_function = np.max
    elif aggregation == 'sum':
        aggregation_function = np.sum
    elif aggregation == 'sqrt':
        aggregation_function = np.sqrt
    else:
        raise ValueError(
            'aggregation only supports `mean`, `min`, `max`, `sum` and `sqrt`'
        )

    vectors = []
    for string in cleaned_strings:
        inside = []
        for token in string.split():
            try:
                inside.append(vectorizer.get_vector_by_name(token))
            except:
                if not soft:
                    pass
                else:
                    arr = np.array(
                        [
                            self._jarowinkler.similarity(token, k)
                            for k in vectorizer.words
                        ]
                    )
                    idx = (-arr).argsort()[0]
                    inside.append(
                        vectorizer.get_vector_by_name(vectorizer.words[idx])
                    )
        vectors.append(aggregation_function(inside, axis = 0))
    similar = cosine_similarity(vectors, vectors)
    similar[similar >= 0.999] = 0
    nx_graph = nx.from_numpy_array(similar)
    scores = nx.pagerank(nx_graph, max_iter = 10000)
    ranked_sentences = sorted(
        ((scores[i], s) for i, s in enumerate(original_strings)), reverse = True
    )
    summary = [r[1] for r in ranked_sentences[:top_k]]
    return ' '.join(summary)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import copy
import json
import math
import re
import six
import tensorflow as tf


def get_shape_list(tensor, expected_rank = None, name = None):
    """Returns a list of the shape of tensor, preferring static dimensions.

	Args:
		tensor: A tf.Tensor object to find the shape of.
		expected_rank: (optional) int. The expected rank of `tensor`. If this is
			specified and the `tensor` has a different rank, and exception will be
			thrown.
		name: Optional name of the tensor for the error message.

	Returns:
		A list of dimensions of the shape of tensor. All static dimensions will
		be returned as python integers, and dynamic dimensions will be returned
		as tf.Tensor scalars.
	"""
    if name is None:
        name = tensor.name

    if expected_rank is not None:
        assert_rank(tensor, expected_rank, name)

    shape = tensor.shape.as_list()

    non_static_indexes = []
    for (index, dim) in enumerate(shape):
        if dim is None:
            non_static_indexes.append(index)

    if not non_static_indexes:
        return shape

    dyn_shape = tf.shape(tensor)
    for index in non_static_indexes:
        shape[index] = dyn_shape[index]
    return shape


def reshape_to_matrix(input_tensor):
    """Reshapes a >= rank 2 tensor to a rank 2 tensor (i.e., a matrix)."""
    ndims = input_tensor.shape.ndims
    if ndims < 2:
        raise ValueError(
            'Input tensor must have at least rank 2. Shape = %s'
            % (input_tensor.shape)
        )
    if ndims == 2:
        return input_tensor

    width = input_tensor.shape[-1]
    output_tensor = tf.reshape(input_tensor, [-1, width])
    return output_tensor


def reshape_from_matrix(output_tensor, orig_shape_list):
    """Reshapes a rank 2 tensor back to its original rank >= 2 tensor."""
    if len(orig_shape_list) == 2:
        return output_tensor

    output_shape = get_shape_list(output_tensor)

    orig_dims = orig_shape_list[0:-1]
    width = output_shape[-1]

    return tf.reshape(output_tensor, orig_dims + [width])


def assert_rank(tensor, expected_rank, name = None):
    """Raises an exception if the tensor rank is not of the expected rank.

	Args:
		tensor: A tf.Tensor to check the rank of.
		expected_rank: Python integer or list of integers, expected rank.
		name: Optional name of the tensor for the error message.

	Raises:
		ValueError: If the expected shape doesn't match the actual shape.
	"""
    if name is None:
        name = tensor.name

    expected_rank_dict = {}
    if isinstance(expected_rank, six.integer_types):
        expected_rank_dict[expected_rank] = True
    else:
        for x in expected_rank:
            expected_rank_dict[x] = True

    actual_rank = tensor.shape.ndims
    if actual_rank not in expected_rank_dict:
        scope_name = tf.get_variable_scope().name
        raise ValueError(
            'For the tensor `%s` in scope `%s`, the actual rank '
            '`%d` (shape = %s) is not equal to the expected rank `%s`'
            % (
                name,
                scope_name,
                actual_rank,
                str(tensor.shape),
                str(expected_rank),
            )
        )


def gather_indexes(sequence_tensor, positions):
    """Gathers the vectors at the specific positions over a minibatch."""
    sequence_shape = get_shape_list(sequence_tensor, expected_rank = 3)
    batch_size = sequence_shape[0]
    seq_length = sequence_shape[1]
    width = sequence_shape[2]

    flat_offsets = tf.reshape(
        tf.range(0, batch_size, dtype = tf.int32) * seq_length, [-1, 1]
    )
    flat_positions = tf.reshape(positions + flat_offsets, [-1])
    flat_sequence_tensor = tf.reshape(
        sequence_tensor, [batch_size * seq_length, width]
    )
    output_tensor = tf.gather(flat_sequence_tensor, flat_positions)
    return output_tensor


# add sequence mask for:
# 1. random shuffle lm modeling---xlnet with random shuffled input
# 2. left2right and right2left language modeling
# 3. conditional generation
def generate_seq2seq_mask(attention_mask, mask_sequence, seq_type, **kargs):
    if seq_type == 'seq2seq':
        if mask_sequence is not None:
            seq_shape = get_shape_list(mask_sequence, expected_rank = 2)
            seq_len = seq_shape[1]
            ones = tf.ones((1, seq_len, seq_len))
            a_mask = tf.matrix_band_part(ones, -1, 0)
            s_ex12 = tf.expand_dims(tf.expand_dims(mask_sequence, 1), 2)
            s_ex13 = tf.expand_dims(tf.expand_dims(mask_sequence, 1), 3)
            a_mask = (1 - s_ex13) * (1 - s_ex12) + s_ex13 * a_mask
            # generate mask of batch x seq_len x seq_len
            a_mask = tf.reshape(a_mask, (-1, seq_len, seq_len))
            out_mask = attention_mask * a_mask
        else:
            ones = tf.ones_like(attention_mask[:1])
            mask = tf.matrix_band_part(ones, -1, 0)
            out_mask = attention_mask * mask
    else:
        out_mask = attention_mask

    return out_mask



"""doc."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import os
import tensorflow as tf
import custom_modeling
import xlnet


def construct_scalar_host_call(
    monitor_dict, model_dir, prefix = '', reduce_fn = None
):
    """
  Construct host calls to monitor training progress on TPUs.
  """

    metric_names = list(monitor_dict.keys())

    def host_call_fn(global_step, *args):
        """actual host call function."""
        step = global_step[0]
        with tf.contrib.summary.create_file_writer(
            logdir = model_dir, filename_suffix = '.host_call'
        ).as_default():
            with tf.contrib.summary.always_record_summaries():
                for i, name in enumerate(metric_names):
                    if reduce_fn is None:
                        scalar = args[i][0]
                    else:
                        scalar = reduce_fn(args[i])
                    with tf.contrib.summary.record_summaries_every_n_global_steps(
                        100, global_step = step
                    ):
                        tf.contrib.summary.scalar(
                            prefix + name, scalar, step = step
                        )

                return tf.contrib.summary.all_summary_ops()

    global_step_tensor = tf.reshape(tf.train.get_or_create_global_step(), [1])
    other_tensors = [tf.reshape(monitor_dict[key], [1]) for key in metric_names]

    return host_call_fn, [global_step_tensor] + other_tensors


def two_stream_loss(FLAGS, features, labels, mems, is_training):
    """Pretraining loss with two-stream attention Transformer-XL."""

    #### Unpack input
    mem_name = 'mems'
    mems = mems.get(mem_name, None)

    inp_k = tf.transpose(features['input_k'], [1, 0])
    inp_q = tf.transpose(features['input_q'], [1, 0])

    seg_id = tf.transpose(features['seg_id'], [1, 0])

    inp_mask = None
    perm_mask = tf.transpose(features['perm_mask'], [1, 2, 0])

    if FLAGS.num_predict is not None:
        # [num_predict x tgt_len x bsz]
        target_mapping = tf.transpose(features['target_mapping'], [1, 2, 0])
    else:
        target_mapping = None

    # target for LM loss
    tgt = tf.transpose(features['target'], [1, 0])

    # target mask for LM loss
    tgt_mask = tf.transpose(features['target_mask'], [1, 0])

    # construct xlnet config and save to model_dir
    xlnet_config = xlnet.XLNetConfig(FLAGS = FLAGS)
    xlnet_config.to_json(os.path.join(FLAGS.model_dir, 'config.json'))

    # construct run config from FLAGS
    run_config = xlnet.create_run_config(is_training, False, FLAGS)

    xlnet_model = xlnet.XLNetModel(
        xlnet_config = xlnet_config,
        run_config = run_config,
        input_ids = inp_k,
        seg_ids = seg_id,
        input_mask = inp_mask,
        mems = mems,
        perm_mask = perm_mask,
        target_mapping = target_mapping,
        inp_q = inp_q,
    )

    output = xlnet_model.get_sequence_output()
    new_mems = {mem_name: xlnet_model.get_new_memory()}
    lookup_table = xlnet_model.get_embedding_table()
    lookup_table_2 = xlnet_model.get_embedding_table2()

    initializer = xlnet_model.get_initializer()

    with tf.variable_scope('model', reuse = tf.AUTO_REUSE):
        # LM loss
        accuracy, lm_loss = custom_modeling.lm_accuracy(
            hidden = output,
            target = tgt,
            n_token = xlnet_config.n_token,
            d_model = xlnet_config.d_model,
            initializer = initializer,
            lookup_table = lookup_table,
            lookup_table_2 = lookup_table_2,
            tie_weight = True,
            bi_data = run_config.bi_data,
            use_tpu = run_config.use_tpu,
        )

    #### Quantity to monitor
    monitor_dict = {}

    if FLAGS.use_bfloat16:
        tgt_mask = tf.cast(tgt_mask, tf.float32)
        lm_loss = tf.cast(lm_loss, tf.float32)
    print(tgt_mask, lm_loss, accuracy)

    total_loss = tf.reduce_sum(lm_loss * tgt_mask) / tf.reduce_sum(tgt_mask)
    #     total_accuracy = tf.reduce_sum(accuracy * tgt_mask) / tf.reduce_sum(
    #         accuracy
    #     )
    total_accuracy = accuracy
    monitor_dict['total_loss'] = total_loss
    monitor_dict['total_accuracy'] = total_accuracy

    return total_loss, total_accuracy, new_mems, monitor_dict


def get_loss(FLAGS, features, labels, mems, is_training):
    """Pretraining loss with two-stream attention Transformer-XL."""
    if FLAGS.use_bfloat16:
        with tf.tpu.bfloat16_scope():
            return two_stream_loss(FLAGS, features, labels, mems, is_training)
    else:
        return two_stream_loss(FLAGS, features, labels, mems, is_training)


def get_classification_loss(FLAGS, features, n_class, is_training):
    """Loss for downstream classification tasks."""

    bsz_per_core = tf.shape(features['input_ids'])[0]

    inp = tf.transpose(features['input_ids'], [1, 0])
    seg_id = tf.transpose(features['segment_ids'], [1, 0])
    inp_mask = tf.transpose(features['input_mask'], [1, 0])
    label = tf.reshape(features['label_ids'], [bsz_per_core])

    xlnet_config = xlnet.XLNetConfig(json_path = FLAGS.model_config_path)
    run_config = xlnet.create_run_config(is_training, True, FLAGS)

    xlnet_model = xlnet.XLNetModel(
        xlnet_config = xlnet_config,
        run_config = run_config,
        input_ids = inp,
        seg_ids = seg_id,
        input_mask = inp_mask,
    )

    summary = xlnet_model.get_pooled_out(
        FLAGS.summary_type, FLAGS.use_summ_proj
    )

    with tf.variable_scope('model', reuse = tf.AUTO_REUSE):

        if FLAGS.cls_scope is not None and FLAGS.cls_scope:
            cls_scope = 'classification_{}'.format(FLAGS.cls_scope)
        else:
            cls_scope = 'classification_{}'.format(FLAGS.task_name.lower())

        per_example_loss, logits = modeling.classification_loss(
            hidden = summary,
            labels = label,
            n_class = n_class,
            initializer = xlnet_model.get_initializer(),
            scope = cls_scope,
            return_logits = True,
        )

        total_loss = tf.reduce_mean(per_example_loss)

        return total_loss, per_example_loss, logits


def get_regression_loss(FLAGS, features, is_training):
    """Loss for downstream regression tasks."""

    bsz_per_core = tf.shape(features['input_ids'])[0]

    inp = tf.transpose(features['input_ids'], [1, 0])
    seg_id = tf.transpose(features['segment_ids'], [1, 0])
    inp_mask = tf.transpose(features['input_mask'], [1, 0])
    label = tf.reshape(features['label_ids'], [bsz_per_core])

    xlnet_config = xlnet.XLNetConfig(json_path = FLAGS.model_config_path)
    run_config = xlnet.create_run_config(is_training, True, FLAGS)

    xlnet_model = xlnet.XLNetModel(
        xlnet_config = xlnet_config,
        run_config = run_config,
        input_ids = inp,
        seg_ids = seg_id,
        input_mask = inp_mask,
    )

    summary = xlnet_model.get_pooled_out(
        FLAGS.summary_type, FLAGS.use_summ_proj
    )

    with tf.variable_scope('model', reuse = tf.AUTO_REUSE):
        per_example_loss, logits = modeling.regression_loss(
            hidden = summary,
            labels = label,
            initializer = xlnet_model.get_initializer(),
            scope = 'regression_{}'.format(FLAGS.task_name.lower()),
            return_logits = True,
        )

        total_loss = tf.reduce_mean(per_example_loss)

        return total_loss, per_example_loss, logits


def get_qa_outputs(FLAGS, features, is_training):
    """Loss for downstream span-extraction QA tasks such as SQuAD."""

    inp = tf.transpose(features['input_ids'], [1, 0])
    seg_id = tf.transpose(features['segment_ids'], [1, 0])
    inp_mask = tf.transpose(features['input_mask'], [1, 0])
    cls_index = tf.reshape(features['cls_index'], [-1])

    seq_len = tf.shape(inp)[0]

    xlnet_config = xlnet.XLNetConfig(json_path = FLAGS.model_config_path)
    run_config = xlnet.create_run_config(is_training, True, FLAGS)

    xlnet_model = xlnet.XLNetModel(
        xlnet_config = xlnet_config,
        run_config = run_config,
        input_ids = inp,
        seg_ids = seg_id,
        input_mask = inp_mask,
    )
    output = xlnet_model.get_sequence_output()
    initializer = xlnet_model.get_initializer()

    return_dict = {}

    # invalid position mask such as query and special symbols (PAD, SEP, CLS)
    p_mask = features['p_mask']

    # logit of the start position
    with tf.variable_scope('start_logits'):
        start_logits = tf.layers.dense(
            output, 1, kernel_initializer = initializer
        )
        start_logits = tf.transpose(tf.squeeze(start_logits, -1), [1, 0])
        start_logits_masked = start_logits * (1 - p_mask) - 1e30 * p_mask
        start_log_probs = tf.nn.log_softmax(start_logits_masked, -1)

    # logit of the end position
    with tf.variable_scope('end_logits'):
        if is_training:
            # during training, compute the end logits based on the
            # ground truth of the start position

            start_positions = tf.reshape(features['start_positions'], [-1])
            start_index = tf.one_hot(
                start_positions, depth = seq_len, axis = -1, dtype = tf.float32
            )
            start_features = tf.einsum('lbh,bl->bh', output, start_index)
            start_features = tf.tile(start_features[None], [seq_len, 1, 1])
            end_logits = tf.layers.dense(
                tf.concat([output, start_features], axis = -1),
                xlnet_config.d_model,
                kernel_initializer = initializer,
                activation = tf.tanh,
                name = 'dense_0',
            )
            end_logits = tf.contrib.layers.layer_norm(
                end_logits, begin_norm_axis = -1
            )

            end_logits = tf.layers.dense(
                end_logits,
                1,
                kernel_initializer = initializer,
                name = 'dense_1',
            )
            end_logits = tf.transpose(tf.squeeze(end_logits, -1), [1, 0])
            end_logits_masked = end_logits * (1 - p_mask) - 1e30 * p_mask
            end_log_probs = tf.nn.log_softmax(end_logits_masked, -1)
        else:
            # during inference, compute the end logits based on beam search

            start_top_log_probs, start_top_index = tf.nn.top_k(
                start_log_probs, k = FLAGS.start_n_top
            )
            start_index = tf.one_hot(
                start_top_index, depth = seq_len, axis = -1, dtype = tf.float32
            )
            start_features = tf.einsum('lbh,bkl->bkh', output, start_index)
            end_input = tf.tile(
                output[:, :, None], [1, 1, FLAGS.start_n_top, 1]
            )
            start_features = tf.tile(start_features[None], [seq_len, 1, 1, 1])
            end_input = tf.concat([end_input, start_features], axis = -1)
            end_logits = tf.layers.dense(
                end_input,
                xlnet_config.d_model,
                kernel_initializer = initializer,
                activation = tf.tanh,
                name = 'dense_0',
            )
            end_logits = tf.contrib.layers.layer_norm(
                end_logits, begin_norm_axis = -1
            )
            end_logits = tf.layers.dense(
                end_logits,
                1,
                kernel_initializer = initializer,
                name = 'dense_1',
            )
            end_logits = tf.reshape(
                end_logits, [seq_len, -1, FLAGS.start_n_top]
            )
            end_logits = tf.transpose(end_logits, [1, 2, 0])
            end_logits_masked = (
                end_logits * (1 - p_mask[:, None]) - 1e30 * p_mask[:, None]
            )
            end_log_probs = tf.nn.log_softmax(end_logits_masked, -1)
            end_top_log_probs, end_top_index = tf.nn.top_k(
                end_log_probs, k = FLAGS.end_n_top
            )
            end_top_log_probs = tf.reshape(
                end_top_log_probs, [-1, FLAGS.start_n_top * FLAGS.end_n_top]
            )
            end_top_index = tf.reshape(
                end_top_index, [-1, FLAGS.start_n_top * FLAGS.end_n_top]
            )

    if is_training:
        return_dict['start_log_probs'] = start_log_probs
        return_dict['end_log_probs'] = end_log_probs
    else:
        return_dict['start_top_log_probs'] = start_top_log_probs
        return_dict['start_top_index'] = start_top_index
        return_dict['end_top_log_probs'] = end_top_log_probs
        return_dict['end_top_index'] = end_top_index

    # an additional layer to predict answerability
    with tf.variable_scope('answer_class'):
        # get the representation of CLS
        cls_index = tf.one_hot(
            cls_index, seq_len, axis = -1, dtype = tf.float32
        )
        cls_feature = tf.einsum('lbh,bl->bh', output, cls_index)

        # get the representation of START
        start_p = tf.nn.softmax(
            start_logits_masked, axis = -1, name = 'softmax_start'
        )
        start_feature = tf.einsum('lbh,bl->bh', output, start_p)

        # note(zhiliny): no dependency on end_feature so that we can obtain
        # one single `cls_logits` for each sample
        ans_feature = tf.concat([start_feature, cls_feature], -1)
        ans_feature = tf.layers.dense(
            ans_feature,
            xlnet_config.d_model,
            activation = tf.tanh,
            kernel_initializer = initializer,
            name = 'dense_0',
        )
        ans_feature = tf.layers.dropout(
            ans_feature, FLAGS.dropout, training = is_training
        )
        cls_logits = tf.layers.dense(
            ans_feature,
            1,
            kernel_initializer = initializer,
            name = 'dense_1',
            use_bias = False,
        )
        cls_logits = tf.squeeze(cls_logits, -1)

        return_dict['cls_logits'] = cls_logits

    return return_dict


def get_race_loss(FLAGS, features, is_training):
    """Loss for downstream multi-choice QA tasks such as RACE."""

    bsz_per_core = tf.shape(features['input_ids'])[0]

    def _transform_features(feature):
        out = tf.reshape(feature, [bsz_per_core, 4, -1])
        out = tf.transpose(out, [2, 0, 1])
        out = tf.reshape(out, [-1, bsz_per_core * 4])
        return out

    inp = _transform_features(features['input_ids'])
    seg_id = _transform_features(features['segment_ids'])
    inp_mask = _transform_features(features['input_mask'])
    label = tf.reshape(features['label_ids'], [bsz_per_core])

    xlnet_config = xlnet.XLNetConfig(json_path = FLAGS.model_config_path)
    run_config = xlnet.create_run_config(is_training, True, FLAGS)

    xlnet_model = xlnet.XLNetModel(
        xlnet_config = xlnet_config,
        run_config = run_config,
        input_ids = inp,
        seg_ids = seg_id,
        input_mask = inp_mask,
    )
    summary = xlnet_model.get_pooled_out(
        FLAGS.summary_type, FLAGS.use_summ_proj
    )

    with tf.variable_scope('logits'):
        logits = tf.layers.dense(
            summary, 1, kernel_initializer = xlnet_model.get_initializer()
        )
        logits = tf.reshape(logits, [bsz_per_core, 4])

        one_hot_target = tf.one_hot(label, 4)
        per_example_loss = -tf.reduce_sum(
            tf.nn.log_softmax(logits) * one_hot_target, -1
        )
        total_loss = tf.reduce_mean(per_example_loss)

    return total_loss, per_example_loss, logits



import numpy as np
import pandas as pd
import pickle, warnings, string

# https://www.kaggle.com/kyakovlev/jigsaw-general-helper-public/code
########################### Url extensions
#################################################################################
url_extensions = """com|net|org|edu|gov|mil|aero|asia|biz|cat|coop|info|int|jobs|mobi|museum|name|post|pro|tel|travel|xxx|nyc|ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|Ja|sk|sl|sm|sn|so|sr|ss|st|su|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|yu|za|zm|zw"""
url_extensions = [c for c in url_extensions.split('|')]


########################### HTML valid tags
#################################################################################
html_tags = [
    'a',
    'abbr',
    'address',
    'area',
    'article',
    'aside',
    'audio',
    'b',
    'base',
    'bdi',
    'bdo',
    'blockquote',
    'body',
    'br',
    'button',
    'canvas',
    'caption',
    'cite',
    'code',
    'col',
    'colgroup',
    'data',
    'datalist',
    'dd',
    'del',
    'details',
    'dfn',
    'dialog',
    'div',
    'dl',
    'dt',
    'em',
    'embed',
    'fieldset',
    'figcaption',
    'figure',
    'footer',
    'form',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6',
    'head',
    'header',
    'hgroup',
    'hr',
    'html',
    'i',
    'iframe',
    'img',
    'input',
    'ins',
    'kbd',
    'keygen',
    'label',
    'legend',
    'li',
    'link',
    'main',
    'map',
    'mark',
    'math',
    'menu',
    'menuitem',
    'meta',
    'meter',
    'nav',
    'noscript',
    'object',
    'ol',
    'optgroup',
    'option',
    'output',
    'p',
    'param',
    'picture',
    'pre',
    'progress',
    'q',
    'rb',
    'rp',
    'rt',
    'rtc',
    'ruby',
    's',
    'samp',
    'script',
    'section',
    'select',
    'slot',
    'small',
    'source',
    'span',
    'strong',
    'style',
    'sub',
    'summary',
    'sup',
    'svg',
    'table',
    'tbody',
    'td',
    'template',
    'textarea',
    'tfoot',
    'th',
    'thead',
    'time',
    'title',
    'tr',
    'track',
    'u',
    'ul',
    'var',
    'video',
    'wbr',
]


########################### Good chars Dieter Kernel
#################################################################################
good_chars_dieter = '.,?!-;*"…:—()%#$&_/@＼・ω+=”“[]^–>\\°<~•≠™ˈʊɒ∞§{}·τα❤☺ɡ|¢→̶`❥━┣┫┗Ｏ►★©―ɪ✔®\x96\x92●£♥➤´¹☕≈÷♡◐║▬′ɔː€۩۞†μ✒➥═☆ˌ◄½ʻπδηλσερνʃ✬ＳＵＰＥＲＩＴ☻±♍µº¾✓◾؟．⬅℅»Вав❣⋅¿¬♫ＣＭβ█▓▒░⇒⭐›¡₂₃❧▰▔◞▀▂▃▄▅▆▇↙γ̄″☹➡«φ⅓„✋：¥̲̅́∙‛◇✏▷❓❗¶˚˙）сиʿ✨。ɑ\x80◕！％¯−ﬂﬁ₁²ʌ¼⁴⁄₄⌠♭✘╪▶☭✭♪☔☠♂☃☎✈✌✰❆☙○‣⚓年∎ℒ▪▙☏⅛ｃａｓǀ℮¸ｗ‚∼‖ℳ❄←☼⋆ʒ⊂、⅔¨͡๏⚾⚽Φ×θ￦？（℃⏩☮⚠月✊❌⭕▸■⇌☐☑⚡☄ǫ╭∩╮，例＞ʕɐ̣Δ₀✞┈╱╲▏▕┃╰▊▋╯┳┊≥☒↑☝ɹ✅☛♩☞ＡＪＢ◔◡↓♀⬆̱ℏ\x91⠀ˤ╚↺⇤∏✾◦♬³の｜／∵∴√Ω¤☜▲↳▫‿⬇✧ｏｖｍ－２０８＇‰≤∕ˆ⚜☁'


########################### Bad chars Dieter Kernel
#################################################################################
bad_chars_dieter = '\n🍕\r🐵😑\xa0\ue014\t\uf818\uf04a\xad😢🐶️\uf0e0😜😎👊\u200b\u200e😁عدويهصقأناخلىبمغر😍💖💵Е👎😀😂\u202a\u202c🔥😄🏻💥ᴍʏʀᴇɴᴅᴏᴀᴋʜᴜʟᴛᴄᴘʙғᴊᴡɢ😋👏שלוםבי😱‼\x81エンジ故障\u2009🚌ᴵ͞🌟😊😳😧🙀😐😕\u200f👍😮😃😘אעכח💩💯⛽🚄🏼ஜ😖ᴠ🚲‐😟😈💪🙏🎯🌹😇💔😡\x7f👌ἐὶήιὲκἀίῃἴξ🙄Ｈ😠\ufeff\u2028😉😤⛺🙂\u3000تحكسة👮💙فزط😏🍾🎉😞\u2008🏾😅😭👻😥😔😓🏽🎆🍻🍽🎶🌺🤔😪\x08‑🐰🐇🐱🙆😨🙃💕𝘊𝘦𝘳𝘢𝘵𝘰𝘤𝘺𝘴𝘪𝘧𝘮𝘣💗💚地獄谷улкнПоАН🐾🐕😆ה🔗🚽歌舞伎🙈😴🏿🤗🇺🇸мυтѕ⤵🏆🎃😩\u200a🌠🐟💫💰💎эпрд\x95🖐🙅⛲🍰🤐👆🙌\u2002💛🙁👀🙊🙉\u2004ˢᵒʳʸᴼᴷᴺʷᵗʰᵉᵘ\x13🚬🤓\ue602😵άοόςέὸתמדףנרךצט😒͝🆕👅👥👄🔄🔤👉👤👶👲🔛🎓\uf0b7\uf04c\x9f\x10成都😣⏺😌🤑🌏😯ех😲Ἰᾶὁ💞🚓🔔📚🏀👐\u202d💤🍇\ue613小土豆🏡❔⁉\u202f👠》कर्मा🇹🇼🌸蔡英文🌞🎲レクサス😛外国人关系Сб💋💀🎄💜🤢َِьыгя不是\x9c\x9d🗑\u2005💃📣👿༼つ༽😰ḷЗз▱ц￼🤣卖温哥华议会下降你失去所有的钱加拿大坏税骗子🐝ツ🎅\x85🍺آإشء🎵🌎͟ἔ油别克🤡🤥😬🤧й\u2003🚀🤴ʲшчИОРФДЯМюж😝🖑ὐύύ特殊作戦群щ💨圆明园קℐ🏈😺🌍⏏ệ🍔🐮🍁🍆🍑🌮🌯🤦\u200d𝓒𝓲𝓿𝓵안영하세요ЖљКћ🍀😫🤤ῦ我出生在了可以说普通话汉语好极🎼🕺🍸🥂🗽🎇🎊🆘🤠👩🖒🚪天一家⚲\u2006⚭⚆⬭⬯⏖新✀╌🇫🇷🇩🇪🇮🇬🇧😷🇨🇦ХШ🌐\x1f杀鸡给猴看ʁ𝗪𝗵𝗲𝗻𝘆𝗼𝘂𝗿𝗮𝗹𝗶𝘇𝗯𝘁𝗰𝘀𝘅𝗽𝘄𝗱📺ϖ\u2000үսᴦᎥһͺ\u2007հ\u2001ɩｙｅ൦ｌƽｈ𝐓𝐡𝐞𝐫𝐮𝐝𝐚𝐃𝐜𝐩𝐭𝐢𝐨𝐧Ƅᴨןᑯ໐ΤᏧ௦Іᴑ܁𝐬𝐰𝐲𝐛𝐦𝐯𝐑𝐙𝐣𝐇𝐂𝐘𝟎ԜТᗞ౦〔Ꭻ𝐳𝐔𝐱𝟔𝟓𝐅🐋ﬃ💘💓ё𝘥𝘯𝘶💐🌋🌄🌅𝙬𝙖𝙨𝙤𝙣𝙡𝙮𝙘𝙠𝙚𝙙𝙜𝙧𝙥𝙩𝙪𝙗𝙞𝙝𝙛👺🐷ℋ𝐀𝐥𝐪🚶𝙢Ἱ🤘ͦ💸ج패티Ｗ𝙇ᵻ👂👃ɜ🎫\uf0a7БУі🚢🚂ગુજરાતીῆ🏃𝓬𝓻𝓴𝓮𝓽𝓼☘﴾̯﴿₽\ue807𝑻𝒆𝒍𝒕𝒉𝒓𝒖𝒂𝒏𝒅𝒔𝒎𝒗𝒊👽😙\u200cЛ‒🎾👹⎌🏒⛸公寓养宠物吗🏄🐀🚑🤷操美𝒑𝒚𝒐𝑴🤙🐒欢迎来到阿拉斯ספ𝙫🐈𝒌𝙊𝙭𝙆𝙋𝙍𝘼𝙅ﷻ🦄巨收赢得白鬼愤怒要买额ẽ🚗🐳𝟏𝐟𝟖𝟑𝟕𝒄𝟗𝐠𝙄𝙃👇锟斤拷𝗢𝟳𝟱𝟬⦁マルハニチロ株式社⛷한국어ㄸㅓ니͜ʖ𝘿𝙔₵𝒩ℯ𝒾𝓁𝒶𝓉𝓇𝓊𝓃𝓈𝓅ℴ𝒻𝒽𝓀𝓌𝒸𝓎𝙏ζ𝙟𝘃𝗺𝟮𝟭𝟯𝟲👋🦊多伦🐽🎻🎹⛓🏹🍷🦆为和中友谊祝贺与其想象对法如直接问用自己猜本传教士没积唯认识基督徒曾经让相信耶稣复活死怪他但当们聊些政治题时候战胜因圣把全堂结婚孩恐惧且栗谓这样还♾🎸🤕🤒⛑🎁批判检讨🏝🦁🙋😶쥐스탱트뤼도석유가격인상이경제황을렵게만들지않록잘관리해야합다캐나에서대마초와화약금의품런성분갈때는반드시허된사용🔫👁凸ὰ💲🗯𝙈Ἄ𝒇𝒈𝒘𝒃𝑬𝑶𝕾𝖙𝖗𝖆𝖎𝖌𝖍𝖕𝖊𝖔𝖑𝖉𝖓𝖐𝖜𝖞𝖚𝖇𝕿𝖘𝖄𝖛𝖒𝖋𝖂𝕴𝖟𝖈𝕸👑🚿💡知彼百\uf005𝙀𝒛𝑲𝑳𝑾𝒋𝟒😦𝙒𝘾𝘽🏐𝘩𝘨ὼṑ𝑱𝑹𝑫𝑵𝑪🇰🇵👾ᓇᒧᔭᐃᐧᐦᑳᐨᓃᓂᑲᐸᑭᑎᓀᐣ🐄🎈🔨🐎🤞🐸💟🎰🌝🛳点击查版🍭𝑥𝑦𝑧ＮＧ👣\uf020っ🏉ф💭🎥Ξ🐴👨🤳🦍\x0b🍩𝑯𝒒😗𝟐🏂👳🍗🕉🐲چی𝑮𝗕𝗴🍒ꜥⲣⲏ🐑⏰鉄リ事件ї💊「」\uf203\uf09a\uf222\ue608\uf202\uf099\uf469\ue607\uf410\ue600燻製シ虚偽屁理屈Г𝑩𝑰𝒀𝑺🌤𝗳𝗜𝗙𝗦𝗧🍊ὺἈἡχῖΛ⤏🇳𝒙ψՁմեռայինրւդձ冬至ὀ𝒁🔹🤚🍎𝑷🐂💅𝘬𝘱𝘸𝘷𝘐𝘭𝘓𝘖𝘹𝘲𝘫کΒώ💢ΜΟΝΑΕ🇱♲𝝈↴💒⊘Ȼ🚴🖕🖤🥘📍👈➕🚫🎨🌑🐻𝐎𝐍𝐊𝑭🤖🎎😼🕷ｇｒｎｔｉｄｕｆｂｋ𝟰🇴🇭🇻🇲𝗞𝗭𝗘𝗤👼📉🍟🍦🌈🔭《🐊🐍\uf10aლڡ🐦\U0001f92f\U0001f92a🐡💳ἱ🙇𝗸𝗟𝗠𝗷🥜さようなら🔼'


########################### Contractions
#################################################################################
contractions = {
    "aren't": 'are not',
    "Aren't": 'Are not',
    "AREN'T": 'ARE NOT',
    "C'est": "C'est",
    "C'mon": "C'mon",
    "c'mon": "c'mon",
    "can't": 'cannot',
    "Can't": 'Cannot',
    "CAN'T": 'CANNOT',
    "con't": 'continued',
    "cont'd": 'continued',
    "could've": 'could have',
    "couldn't": 'could not',
    "Couldn't": 'Could not',
    "didn't": 'did not',
    "Didn't": 'Did not',
    "DIDN'T": 'DID NOT',
    "don't": 'do not',
    "Don't": 'Do not',
    "DON'T": 'DO NOT',
    "doesn't": 'does not',
    "Doesn't": 'Does not',
    "else's": 'else',
    "gov's": 'government',
    "Gov's": 'government',
    "gov't": 'government',
    "Gov't": 'government',
    "govt's": 'government',
    "gov'ts": 'governments',
    "hadn't": 'had not',
    "hasn't": 'has not',
    "Hasn't": 'Has not',
    "haven't": 'have not',
    "Haven't": 'Have not',
    "he's": 'he is',
    "He's": 'He is',
    "he'll": 'he will',
    "He'll": 'He will',
    "he'd": 'he would',
    "He'd": 'He would',
    "Here's": 'Here is',
    "here's": 'here is',
    "I'm": 'I am',
    "i'm": 'i am',
    "I'M": 'I am',
    "I've": 'I have',
    "i've": 'i have',
    "I'll": 'I will',
    "i'll": 'i will',
    "I'd": 'I would',
    "i'd": 'i would',
    "ain't": 'is not',
    "isn't": 'is not',
    "Isn't": 'Is not',
    "ISN'T": 'IS NOT',
    "it's": 'it is',
    "It's": 'It is',
    "IT'S": 'IT IS',
    "I's": 'It is',
    "i's": 'it is',
    "it'll": 'it will',
    "It'll": 'It will',
    "it'd": 'it would',
    "It'd": 'It would',
    "Let's": "Let's",
    "let's": 'let us',
    "ma'am": 'madam',
    "Ma'am": 'Madam',
    "she's": 'she is',
    "She's": 'She is',
    "she'll": 'she will',
    "She'll": 'She will',
    "she'd": 'she would',
    "She'd": 'She would',
    "shouldn't": 'should not',
    "that's": 'that is',
    "That's": 'That is',
    "THAT'S": 'THAT IS',
    "THAT's": 'THAT IS',
    "that'll": 'that will',
    "That'll": 'That will',
    "there's": 'there is',
    "There's": 'There is',
    "there'll": 'there will',
    "There'll": 'There will',
    "there'd": 'there would',
    "they're": 'they are',
    "They're": 'They are',
    "they've": 'they have',
    "They've": 'They Have',
    "they'll": 'they will',
    "They'll": 'They will',
    "they'd": 'they would',
    "They'd": 'They would',
    "wasn't": 'was not',
    "we're": 'we are',
    "We're": 'We are',
    "we've": 'we have',
    "We've": 'We have',
    "we'll": 'we will',
    "We'll": 'We will',
    "we'd": 'we would',
    "We'd": 'We would',
    "What'll": 'What will',
    "weren't": 'were not',
    "Weren't": 'Were not',
    "what's": 'what is',
    "What's": 'What is',
    "When's": 'When is',
    "Where's": 'Where is',
    "where's": 'where is',
    "Where'd": 'Where would',
    "who're": 'who are',
    "who've": 'who have',
    "who's": 'who is',
    "Who's": 'Who is',
    "who'll": 'who will',
    "who'd": 'Who would',
    "Who'd": 'Who would',
    "won't": 'will not',
    "Won't": 'will not',
    "WON'T": 'WILL NOT',
    "would've": 'would have',
    "wouldn't": 'would not',
    "Wouldn't": 'Would not',
    "would't": 'would not',
    "Would't": 'Would not',
    "y'all": 'you all',
    "Y'all": 'You all',
    "you're": 'you are',
    "You're": 'You are',
    "YOU'RE": 'YOU ARE',
    "you've": 'you have',
    "You've": 'You have',
    "y'know": 'you know',
    "Y'know": 'You know',
    "ya'll": 'you will',
    "you'll": 'you will',
    "You'll": 'You will',
    "you'd": 'you would',
    "You'd": 'You would',
    "Y'got": 'You got',
    'cause': 'because',
    "had'nt": 'had not',
    "Had'nt": 'Had not',
    "how'd": 'how did',
    "how'd'y": 'how do you',
    "how'll": 'how will',
    "how's": 'how is',
    "I'd've": 'I would have',
    "I'll've": 'I will have',
    "i'd've": 'i would have',
    "i'll've": 'i will have',
    "it'd've": 'it would have',
    "it'll've": 'it will have',
    "mayn't": 'may not',
    "might've": 'might have',
    "mightn't": 'might not',
    "mightn't've": 'might not have',
    "must've": 'must have',
    "mustn't": 'must not',
    "mustn't've": 'must not have',
    "needn't": 'need not',
    "needn't've": 'need not have',
    "o'clock": 'of the clock',
    "oughtn't": 'ought not',
    "oughtn't've": 'ought not have',
    "shan't": 'shall not',
    "sha'n't": 'shall not',
    "shan't've": 'shall not have',
    "she'd've": 'she would have',
    "she'll've": 'she will have',
    "should've": 'should have',
    "shouldn't've": 'should not have',
    "so've": 'so have',
    "so's": 'so as',
    "this's": 'this is',
    "that'd": 'that would',
    "that'd've": 'that would have',
    "there'd've": 'there would have',
    "they'd've": 'they would have',
    "they'll've": 'they will have',
    "to've": 'to have',
    "we'd've": 'we would have',
    "we'll've": 'we will have',
    "what'll": 'what will',
    "what'll've": 'what will have',
    "what're": 'what are',
    "what've": 'what have',
    "when's": 'when is',
    "when've": 'when have',
    "where'd": 'where did',
    "where've": 'where have',
    "who'll've": 'who will have',
    "why's": 'why is',
    "why've": 'why have',
    "will've": 'will have',
    "won't've": 'will not have',
    "wouldn't've": 'would not have',
    "y'all'd": 'you all would',
    "y'all'd've": 'you all would have',
    "y'all're": 'you all are',
    "y'all've": 'you all have',
    "you'd've": 'you would have',
    "you'll've": 'you will have',
}


########################### Chars Normalization
#################################################################################
normalized_chars = {}

chars = '‒–―‐—━—-▬'
for char in chars:
    normalized_chars[ord(char)] = '-'

chars = '«»“”¨"'
for char in chars:
    normalized_chars[ord(char)] = '"'

chars = "’'ʻˈ´`′‘’\x92"
for char in chars:
    normalized_chars[ord(char)] = "'"

chars = '̲_'
for char in chars:
    normalized_chars[ord(char)] = '_'

chars = '\xad\x7f'
for char in chars:
    normalized_chars[ord(char)] = ''

chars = '\n\r\t\u200b\x96'
for char in chars:
    normalized_chars[ord(char)] = ' '


########################### White list chars
#################################################################################
latin_similar = "’'‘ÆÐƎƏƐƔĲŊŒẞÞǷȜæðǝəɛɣĳŋœĸſßþƿȝĄƁÇĐƊĘĦĮƘŁØƠŞȘŢȚŦŲƯY̨Ƴąɓçđɗęħįƙłøơşșţțŧųưy̨ƴÁÀÂÄǍĂĀÃÅǺĄÆǼǢƁĆĊĈČÇĎḌĐƊÐÉÈĖÊËĚĔĒĘẸƎƏƐĠĜǦĞĢƔáàâäǎăāãåǻąæǽǣɓćċĉčçďḍđɗðéèėêëěĕēęẹǝəɛġĝǧğģɣĤḤĦIÍÌİÎÏǏĬĪĨĮỊĲĴĶƘĹĻŁĽĿʼNŃN̈ŇÑŅŊÓÒÔÖǑŎŌÕŐỌØǾƠŒĥḥħıíìiîïǐĭīĩįịĳĵķƙĸĺļłľŀŉńn̈ňñņŋóòôöǒŏōõőọøǿơœŔŘŖŚŜŠŞȘṢẞŤŢṬŦÞÚÙÛÜǓŬŪŨŰŮŲỤƯẂẀŴẄǷÝỲŶŸȲỸƳŹŻŽẒŕřŗſśŝšşșṣßťţṭŧþúùûüǔŭūũűůųụưẃẁŵẅƿýỳŷÿȳỹƴźżžẓ"
white_list_chars = string.ascii_letters + string.digits + latin_similar + ' '
white_list_chars += "'"


########################### Text Pictograms / Emotions
#################################################################################
pictograms = [
    [
        [
            ':‑)',
            ';-))',
            ':)',
            ':-]',
            ':]',
            ':-3',
            ':3',
            ':->',
            ':>',
            '8-)',
            '8)',
            ':-}',
            ':}',
            ':o)',
            ':c)',
            ':^)',
            '=]',
            '=)',
            ':)]',
            ';-}',
            ':>)',
            ':<)',
            ':~)',
            ':))',
            ':-]',
            '=)',
            ':).',
            ':),',
            ':-}',
            ':-)',
        ],
        ['😁'],
    ],
    [
        [':‑D', ':D', '8‑D', '8D', 'x‑D', 'xD', 'X‑D', 'XD', '=D', '=3', 'B^D'],
        ['😃'],
    ],
    [[':-))'], ['😃']],
    [
        [
            ':‑(',
            ':(',
            ':‑c',
            ':c',
            ':‑<',
            ':<',
            ':‑[',
            ':[',
            ':-||',
            '>:[',
            ':{',
            ':@',
            '>:(',
            ':(.',
            '>_<',
            ':-(',
            ':0(',
        ],
        ['😡'],
    ],
    [[":'‑(", ":'(", ':,('], ['😢']],
    [[":'‑)", ":')"], ['😂']],
    [["D‑':", 'D:<', 'D:', 'D8', 'D;', 'D=', 'DX'], ['😱']],
    [
        [
            ':‑O',
            ':O',
            ':‑o',
            ':o',
            ':-0',
            '8‑0',
            '>:O',
            '=:o',
            ':o)',
            ':0)',
            ':o)]',
            '=o)',
        ],
        ['😮'],
    ],
    [[':-*', ':*', ':×'], ['😘']],
    [
        [
            ';‑)',
            ';)',
            '*-)',
            '*)',
            ';‑]',
            ';]',
            ';^)',
            ':‑,',
            ';D',
            ';-))',
            ';>)',
            '=)~',
        ],
        ['😜'],
    ],
    [
        [
            ':‑P',
            ':P',
            'X‑P',
            'XP',
            'x‑p',
            'xp',
            ':‑p',
            ':p',
            ':‑Þ',
            ':Þ',
            ':‑þ',
            ':þ',
            ':‑b',
            ':b',
            'd:',
            '=p',
            '>:P',
            '>:/',
            ':)"',
        ],
        ['😛'],
    ],
    [
        [
            ':‑/',
            ':/',
            ':‑.',
            '>:\\',
            '>:/',
            ':\\',
            '=/',
            '=\\',
            ':L',
            '=L',
            ':S',
        ],
        ['🤔'],
    ],
    [[':‑|', ':|', ':-|', ';-|', ':|'], ['😐']],
    [[':$'], ['😞']],
    [[':‑X', ':X', ':‑#', ':#', ':‑&', ':&'], ['🤐']],
    [['O:‑)', 'O:)', '0:‑3', '0:3', '0:‑)', '0:)', '0;^)'], ['😇']],
    [['>:‑)', '>:)', '}:‑)', '}:)', '3:‑)', '3:)', '>;)'], ['😈']],
    [['|;‑)', '|‑O'], ['😎']],
    [[':‑J'], ['😏']],
    [['%‑)', '%)'], ['😵']],
    [[':‑###..', ':###..'], ['🤢']],
    [['@};- ', '@}->-- ', "@}‑;‑'‑‑‑", '@>‑‑>‑‑'], ['🌹']],
    [['5:‑)', '~:‑\\'], ['Elvis']],
    [['*<|:‑)'], ['🎅']],
    [['~(_8^(I)'], ['Homer Simpson']],
    [['=:o]'], ['Bill Clinton']],
    [['7:^]', ',:‑)'], ['Ronald Reagan']],
    [['</3', '<\3'], ['💔']],
    [['<3'], ['❤']],
    [['><>', '<*)))‑{', '><(((*>'], ['Fish']],
    [['\o/'], ['Yay, yay']],
    [['*\0/*'], ['Cheerleader']],
    [['//0‑0\\\\'], ['John Lennon']],
    [['v.v'], ['😱']],
    [
        ['O_O', 'o‑o', 'O_o', 'o_O', 'o_o', 'O-O', 'o_x', '0_x', 'O_o', 'x_x'],
        ['😮'],
    ],
    [['>.<'], ['🤔']],
]

pictograms_to_emoji = {}
for line in pictograms:
    for k in line[0]:
        pictograms_to_emoji[k] = line[1][0]

########################### Toxic words
#################################################################################
toxic_misspell_dict = {
    's*it': 'shit',
    'sh*t': 'shit',
    's**t': 'shit',
    'shi***': 'shitty',
    'shi**y': 'shitty',
    'shi*': 'shit',
    'shi*s': 'shits',
    'dipsh*t': 'dipshit',
    's%#*': 'shit',
    'Bu**sh*t': 'Bullshit',
    'bullsh*t': 'bullshit',
    'b******t': 'bullshit',
    'b*llsh*t': 'bullshit',
    'batsh*t': 'batshit',
    'sh*tbird': 'shitbird',
    'sh*tty': 'shitty',
    'bullsh*tter': 'bullshitter',
    'sh@#': 'shit',
    'Sh^t': 'shit',
    'sh^t': 'shit',
    'sh@t': 'shit',
    '$het': 'Shit',
    '$h!t': 'shit',
    'sh1t': 'shit',
    'shlt': 'shit',
    '$h*t': 'shit',
    'bat-s**t': 'bat-shit',
    '$hit': 'shit',
    'sh!te': 'shit',
    'sh!t': 'shit',
    'bullsh1t': 'bullshit',
    'b...llsh1t': 'bullshit',
    's***': 'suck',
    '5h1t': 'shit',
    'sh*thole': 'shithole',
    'bats**t': 'batshit',
    'S**t': 'shit',
    'Batsh*t': 'batshit',
    'Bullsh*t': 'Bullshit',
    'SH*T': 'shit',
    'sh**t': 'shit',
    'sh*t-kicked': 'shit kicked',
    's@!t': 'shit',
    'sh@%%t': 'shit',
    's#@t': 'shit',
    '#%@$': 'shit',
    '#*@$': 'shit',
    '^@%#': 'fuck',
    '!@#$': 'fuck',
    's@#t': 'shit',
    'sh@tless': 'shitless',
    '8&@!': 'shit',
    '!@#!!': 'shit',
    '$@%t': 'shit',
    'f@$#': 'fuck',
    'F$$@': 'fuck',
    'F@#$': 'fuck',
    '#@!!%': 'fuck',
    '$%@%ing': 'fucking',
    'F@#k': 'fuck',
    'f@#k': 'fuck',
    '#@!&*': 'fuck',
    'f@#&ing': 'fucking',
    'f@#$^*': 'fucked',
    'F$#@ing': 'fucking',
    '#!$@%^': 'fuck',
    'FU!@#': 'fuck',
    '#%$@er': 'fucker',
    '$@%ing': 'fucking',
    '%#@&ing': 'fucking',
    "*&@k's": 'fuckers',
    '!@#$$!': 'fuck',
    'F#@%ing': 'fucking',
    'F#@*': 'fuck',
    'f@!k': 'fuck',
    '*^@$': 'fuck',
    '$/#@': 'fuck',
    'F!@#in': 'fucking',
    'Fuc@ed': 'fucked',
    'fu@&d': 'fucked',
    'F&@&': 'fuck',
    '$#@^': 'fuck',
    '&@$#': 'fuck',
    '$%@!': 'fuck',
    'fu@#$&': 'fucked',
    '*@#k': 'fuck',
    'F@%!s': 'fucks',
    'fuc$&@ng': 'fucking',
    'f#@k': 'fuck',
    '!$@#%': 'fuck',
    'f******': 'fucking',
    ' f***ing': 'fucking',
    'motherf***ing': 'motherfucking',
    'f***': 'fuck',
    'f***ked': 'fucked',
    'f***ed': 'fucked',
    'fu**ing': 'fucking',
    'clusterf*ck': 'clusterfuck',
    'ratf*cking': 'ratfucking',
    'f*ck': 'fuck',
    'f**k': 'fuck',
    "f**kin'": 'fucking',
    'F**K': 'fuck',
    'F***': 'fuck',
    'F*ck': 'fuck',
    'f**ks': 'fuck',
    'f**cker': 'fucker',
    'F******': 'fucked',
    'f*&$ing': 'fucking',
    'f*k': 'fuck',
    'F*ggot': 'faggot',
    'F*cks': 'fucks',
    'F*CKING': 'fucking',
    'F*** O**': 'fuck off',
    'f*** o**': 'fuck off',
    'f-up': 'fuck up',
    'F-up': 'fuck up',
    'F@#@CK': 'fuck',
    'F---ck': 'fuck',
    'f---ck': 'fuck',
    'f--ck': 'fuck',
    'F--ck': 'fuck',
    'f-ck': 'fuck',
    'F-ck': 'fuck',
    'f-ckin': 'fucking',
    'fu#$ed': 'fucked',
    'f*$(': 'fuck',
    ' f*$K': 'fuck',
    'f__k': 'fuck',
    'f.ck': 'fuck',
    'fck': 'fuck',
    'Fck': 'fuck',
    'F*ing': 'fucking',
    'f*ing': 'fucking',
    'fukin': 'fucking',
    'fuking': 'fucking',
    'f++k': 'fuck',
    'f*%k': 'fuck',
    '.uck': 'fuck',
    'F@ck': 'fuck',
    'fcuking': 'fucking',
    'a55es': 'asses',
    'a**': 'ass',
    'a*#': 'ass',
    'a******': 'asshole',
    'a*****e': 'asshole',
    '@ss': 'ass',
    '@$$': 'ass',
    'A**': 'ass',
    'A**hole': 'asshole',
    '@##': 'ass',
    '@#$': 'ass',
    'a-hole': 'asshole',
    '@sshole': 'asshole',
    '@ssholes': 'asshole',
    'A@@': 'ass',
    'a!@$#$ed': 'assed',
    'ass@s': 'asses',
    'a@#': 'ass',
    'AS^*$@$': 'asses',
    'A#@#$': 'asses',
    '@&&': 'ass',
    'b!tch': 'bitch',
    'b1tch': 'bitch',
    'b*tch': 'bitch',
    'b***h': 'bitch',
    'b***s': 'bitchs',
    'b*th': 'bitch',
    'bit*#^s': 'bitch',
    'b*tt': 'butt',
    'B****': 'bitch',
    'Bit@#$': 'bitch',
    'B***h': 'bitch',
    'Bit*h': 'bitch',
    'bit*h': 'bitch',
    'b****': 'bitch',
    'Bi^@h': 'bitch',
    'B@##S': 'bitchs',
    'Bat-h': 'bitch',
    'b@##$': 'bitch',
    'B@##s': 'bitchs',
    'bit@$': 'bitch',
    'b!t@h': 'bitch',
    'dumb***es': 'dumbasses',
    'Dumb*ss': 'Dumbass',
    'dumba*ss': 'dumbass',
    'broke-a**': 'broke-ass',
    'a***oles': 'assholes',
    'a**holes': 'assholes',
    'da*ned': 'damned',
    'c*#ksukking': 'cock sucking',
    'c***': 'cock',
    'p***y': 'putty',
    'p****': 'putty',
    'P***Y': 'pussy',
    'p***y-grabbing': 'pussy-grabbing',
    'p@$$y': 'pussy',
    'pu$$y': 'pussy',
    'pus$y': 'pussy',
    'pu$sy': 'pussy',
    'p*ssy': 'pussy',
    'pu@#y': 'pussy',
    'p@#$y': 'pussy',
    'puXXy': 'puxxy',
    'puxxy': 'puxxy',
    'N***ga': 'Nigga',
    's*ck': 'suck',
    'suckees': 'sucker',
    'suckee': 'sucker',
    's@#k': 'suck',
    's%@': 'suck',
    's@#K': 'suck',
    'd#$k': 'dick',
    'd@#K': 'dick',
    'd@mn': 'damn',
    'D@mn': 'damn',
    'D@MN': 'damn',
    'da@m': 'damn',
    'p0rn': 'porn',
    '$ex': 'sex',
    'b@stard': 'bastard',
    'b@st@rd': 'bastard',
    'b@#$%^&s': 'bastards',
    'bast@#ds': 'bastards',
    'bas#$@d': 'bastard',
    'b@ssturds': 'bastards',
    'stu*pid': 'stupid',
    'F@KE': 'fake',
    'F@ke': 'fake',
    'N#$@#er': 'nutshell',
    '1%ers': 'very rich people',
    'f@rt': 'fart',
    'd00d': 'dude',
    'n00b': 'noob',
    'ret@rd$': 'retards',
    'ret@rd': 'retard',
    'id-iot': 'idiot',
    'chickens**ts': 'chickenshat',
    'chickens**t': 'chickenshat',
    '0bama': 'obama',
    'ofass': 'of ass',
    'b@t': 'bat',
    'cr@p': 'crap',
    'kr@p': 'crap',
    'c&@p': 'crap',
    'kr@ppy': 'crappy',
    'wh@re': 'whore',
    'b@ll': 'ball',
    'b@ll$': 'balls',
    '6@!!': 'ball',
    'r@pe': 'rape',
    'f@ggot': 'faggot',
    '#@$%': 'cock',
    'su@k': 'suck',
    'r@cist': 'racist',
    'r@ce': 'race',
    'h@ll': 'hell',
    'Isl@m': 'islam',
    '$@rew': 'screwed',
    'scr@wed': 'screwed',
    'j@rk': 'jark',
    's@x': 'sex',
    'idi@t': 'idiot',
    'r@ping': 'raping',
    'V@gina': 'virgina',
    'P^##@*': 'pissed',
    '$k@nk': 'skank',
    'N@zi': 'nazi',
    'MANIA': 'Make America a Nitwit Idiocracy Again',
    'B@t$h!t': 'batshit',
    'bats@3t': 'batshit',
    'f@g': 'fag',
    'R@pe': 'rape',
    's*#@t': 'slot',
    'p@ssw0rd': 'password',
    'p@assword': 'password',
    'Sh*t': 'shit',
    's**T': 'shit',
    'S**T': 'shit',
    'bullSh*t': 'bullshit',
    'BULLSH*T': 'bullshit',
    'B******T': 'bullshit',
    'Bullsh*tter': 'bullshitter',
    'sh1T': 'shit',
    'Sh1t': 'shit',
    'SH1T': 'shit',
    '$Hit': 'shit',
    '$HIT': 'shit',
    'sh!T': 'shit',
    'Sh!t': 'shit',
    'SH!T': 'shit',
    'Bullsh1t': 'bullshit',
    'S***': 'suck',
    'F***ing': 'fucking',
    'F***ked': 'fucked',
    'Fu**ing': 'fucking',
    'F*CK': 'fuck',
    'F**k': 'fuck',
    'F**ks': 'fuck',
    'F**KS': 'fuck',
    'F*k': 'fuck',
    'F-ckin': 'fucking',
    'F__k': 'fuck',
    'F__K': 'fuck',
    'F.ck': 'fuck',
    'fCk': 'fuck',
    'FcK': 'fuck',
    'FCK': 'fuck',
    'Fukin': 'fucking',
    'f++K': 'fuck',
    'F*%k': 'fuck',
    'A*****e': 'asshole',
    '@SS': 'ass',
    'A-hole': 'asshole',
    'A-Hole': 'asshole',
    'A-HOLE': 'asshole',
    'A@#': 'ass',
    'B!tch': 'bitch',
    'B!TCH': 'bitch',
    'B*tch': 'bitch',
    'B***S': 'bitchs',
    'B*tt': 'butt',
    'DUMBA*SS': 'dumbass',
    'A**holes': 'assholes',
    'A**holeS': 'assholes',
    'C***': 'cock',
    'P***y': 'putty',
    'P****': 'putty',
    'P@$$Y': 'pussy',
    'Pu$$y': 'pussy',
    'PU$$Y': 'pussy',
    'PuS$y': 'pussy',
    'P*ssy': 'pussy',
    'Puxxy': 'puxxy',
    'N00b': 'noob',
    '0Bama': 'obama',
    'B@t': 'bat',
    'Cr@p': 'crap',
    'CR@P': 'crap',
    'Kr@p': 'crap',
    'B@ll': 'ball',
    'P@ssw0rd': 'password',
    'bat****': 'batshit',
    'Bat****': 'batshit',
    'a******s': 'assholes',
    'p****d': 'passed',
    's****': 'shit',
    'S****': 'shit',
    'bull****': 'bullshit',
    'Bull****': 'bullshit',
    'n*****': 'niggar',
    'b*****d': 'bastard',
    'r*****d': 'retarded',
    'f*****g': 'fucking',
    "a******'s": 'asshole',
    'f****': 'fuck',
    'moth******': 'mother fucker',
    'F******g': 'fucking',
    'n****r': 'niggar',
    'cr*p': 'crap',
    'a-holes': 'asshole',
    'f--k': 'fuck',
    'a**hole': 'asshole',
    'a$$': 'ass',
    'a$s': 'ass',
    'as$': 'ass',
    '@$s': 'ass',
    '@s$': 'ass',
    '$h': 'sh',
    'f***ing': 'fucking',
    '*ss': 'ass',
    'h***': 'hell',
    'p---y': 'pussy',
    "f'n": 'fucking',
    '*&^%': 'shit',
    'a$$hole': 'asshole',
    'p**sy': 'pussy',
    'f---': 'fuck',
    'pi$$': 'piss',
    "f'd up": 'fucked up',
    'c**k': 'cock',
    'a**clown': 'assclown',
    'p___y': 'pussy',
    'sh--': 'shit',
    'f.cking': 'fucking',
    'a--': 'ass',
    'N—–': 'nigga',
    's*x': 'sex',
    'notalent@$$clown': 'no talent assclown',
    'f--king': 'fucking',
    'a--hole': 'asshole',
    '#whitefragilitycankissmyass': '# white fragility can kiss my ass',
    'N*****': 'niggar',
    'B*****d': 'bastard',
    'F*****G': 'fucking',
    'F****': 'fuck',
    'N****r': 'niggar',
    'Cr*p': 'crap',
    'A-holes': 'asshole',
    'A-Holes': 'asshole',
    'A-HOLES': 'asshole',
    'F--k': 'fuck',
    'F--K': 'fuck',
    'A$$': 'ass',
    '@$S': 'ass',
    '$H': 'sh',
    'F***ing': 'fucking',
    '*SS': 'ass',
    'H***': 'hell',
    'P---y': 'pussy',
    "F'n": 'fucking',
    "F'N": 'fucking',
    'A$$hole': 'asshole',
    'A$$HOLE': 'asshole',
    'P**sy': 'pussy',
    'P**SY': 'pussy',
    'F---': 'fuck',
    'Pi$$': 'piss',
    "F'd up": 'fucked up',
    "F'D UP": 'fucked up',
    'C**k': 'cock',
    'P___y': 'pussy',
    'Sh--': 'shit',
    'SH--': 'shit',
    'A--': 'ass',
    'S*x': 'sex',
    'F--king': 'fucking',
    'A--HOLE': 'asshole',
    'pi**ing': 'pissing',
    '**ok': 'fuck',
    'bi*ch': 'bitch',
    'Sh*ts': 'Shits',
    'Rat****ing': 'fuck',
    '*ds': 'faggots',
    'C*nt': 'Cunt',
    '***ed': 'assholed',
    'h*ll': 'asshole',
    'Re*****s': 'Retards',
    'c*unt': 'cunt',
    'f*rt': 'fuck',
    'p***ing': 'pissing',
    'Pi**ing': 'Pissing',
    'd**m': 'Damn',
    'f***': 'fuck',
    's*': 'Suck',
    'c*nt': 'cunt',
    'dam*d': 'damned',
    'nigg*r': 'nigger',
    'an*l': 'anal',
    'f**t': 'faggot',
    's***': 'shit',
    'H*ll': 'asshole',
    'p***ed': 'pissed',
    'a**ed': 'assholed',
    'd****d': 'fuck',
    'they*you': 'fuck',
    '*****RG': 'fuck',
    'a*s': 'ass',
    'h**l': 'asshole',
    'a*sholes': 'assholes',
    'b****': 'bitch',
    'd*ck': 'fuck',
    'H**L': 'Asshole',
    'mother*cking': 'mother fucking',
    'b*tch': 'bitch',
    'as**in': 'ass',
    'motherfu**ers': 'mother fuckers',
    'bull**it': 'bullshit',
    '****may': '**** may',
    '*Let': '* Let',
}



# coding=utf-8
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unicodedata
import six
from functools import partial


SPIECE_UNDERLINE = '▁'


def printable_text(text):
    """Returns text encoded in a way suitable for print or `tf.logging`."""

    # These functions want `str` for both Python2 and Python3, but in one case
    # it's a Unicode string and in the other it's a byte string.
    if six.PY3:
        if isinstance(text, str):
            return text
        elif isinstance(text, bytes):
            return text.decode('utf-8', 'ignore')
        else:
            raise ValueError('Unsupported string type: %s' % (type(text)))
    elif six.PY2:
        if isinstance(text, str):
            return text
        elif isinstance(text, unicode):
            return text.encode('utf-8')
        else:
            raise ValueError('Unsupported string type: %s' % (type(text)))
    else:
        raise ValueError('Not running on Python2 or Python 3?')


def print_(*args):
    new_args = []
    for arg in args:
        if isinstance(arg, list):
            s = [printable_text(i) for i in arg]
            s = ' '.join(s)
            new_args.append(s)
        else:
            new_args.append(printable_text(arg))
    print(*new_args)


def preprocess_text(
    inputs, lower = False, remove_space = True, keep_accents = False
):
    if remove_space:
        outputs = ' '.join(inputs.strip().split())
    else:
        outputs = inputs
    outputs = outputs.replace('``', '"').replace("''", '"')

    if six.PY2 and isinstance(outputs, str):
        outputs = outputs.decode('utf-8')

    if not keep_accents:
        outputs = unicodedata.normalize('NFKD', outputs)
        outputs = ''.join([c for c in outputs if not unicodedata.combining(c)])
    if lower:
        outputs = outputs.lower()

    return outputs


def encode_pieces(sp_model, text, return_unicode = True, sample = False):
    # return_unicode is used only for py2

    # note(zhiliny): in some systems, sentencepiece only accepts str for py2
    if six.PY2 and isinstance(text, unicode):
        text = text.encode('utf-8')

    if not sample:
        pieces = sp_model.EncodeAsPieces(text)
    else:
        pieces = sp_model.SampleEncodeAsPieces(text, 64, 0.1)
    new_pieces = []
    for piece in pieces:
        if len(piece) > 1 and piece[-1] == ',' and piece[-2].isdigit():
            cur_pieces = sp_model.EncodeAsPieces(
                piece[:-1].replace(SPIECE_UNDERLINE, '')
            )
            if (
                piece[0] != SPIECE_UNDERLINE
                and cur_pieces[0][0] == SPIECE_UNDERLINE
            ):
                if len(cur_pieces[0]) == 1:
                    cur_pieces = cur_pieces[1:]
                else:
                    cur_pieces[0] = cur_pieces[0][1:]
            cur_pieces.append(piece[-1])
            new_pieces.extend(cur_pieces)
        else:
            new_pieces.append(piece)

    # note(zhiliny): convert back to unicode for py2
    if six.PY2 and return_unicode:
        ret_pieces = []
        for piece in new_pieces:
            if isinstance(piece, str):
                piece = piece.decode('utf-8')
            ret_pieces.append(piece)
        new_pieces = ret_pieces

    return new_pieces


def encode_ids(sp_model, text, sample = False):
    pieces = encode_pieces(
        sp_model, text, return_unicode = False, sample = sample
    )
    ids = [sp_model.PieceToId(piece) for piece in pieces]
    return ids


if __name__ == '__main__':
    import sentencepiece as spm

    sp = spm.SentencePieceProcessor()
    sp.load('sp10m.uncased.v3.model')

    print_(u'I was born in 2000, and this is falsé.')
    print_(
        u'ORIGINAL',
        sp.EncodeAsPieces(u'I was born in 2000, and this is falsé.'),
    )
    print_(
        u'OURS', encode_pieces(sp, u'I was born in 2000, and this is falsé.')
    )
    print(encode_ids(sp, u'I was born in 2000, and this is falsé.'))
    print_('')
    prepro_func = partial(preprocess_text, lower = True)
    print_(prepro_func('I was born in 2000, and this is falsé.'))
    print_(
        'ORIGINAL',
        sp.EncodeAsPieces(
            prepro_func('I was born in 2000, and this is falsé.')
        ),
    )
    print_(
        'OURS',
        encode_pieces(
            sp, prepro_func('I was born in 2000, and this is falsé.')
        ),
    )
    print(encode_ids(sp, prepro_func('I was born in 2000, and this is falsé.')))
    print_('')
    print_('I was born in 2000, and this is falsé.')
    print_(
        'ORIGINAL', sp.EncodeAsPieces('I was born in 2000, and this is falsé.')
    )
    print_('OURS', encode_pieces(sp, 'I was born in 2000, and this is falsé.'))
    print(encode_ids(sp, 'I was born in 2000, and this is falsé.'))
    print_('')
    print_('I was born in 92000, and this is falsé.')
    print_(
        'ORIGINAL', sp.EncodeAsPieces('I was born in 92000, and this is falsé.')
    )
    print_('OURS', encode_pieces(sp, 'I was born in 92000, and this is falsé.'))
    print(encode_ids(sp, 'I was born in 92000, and this is falsé.'))



from ._utils import _softmax_class
from ._utils._paths import PATH_SUBJECTIVE, S3_PATH_SUBJECTIVE

_availability = {
    'bert': ['base', 'small'],
    'xlnet': ['base'],
    'albert': ['base'],
}


def available_transformer_model():
    """
    List available transformer subjective analysis models.
    """
    return _availability


def multinomial(validate = True):
    """
    Load multinomial subjectivity model.

    Parameters
    ----------
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    BAYES : malaya._models._sklearn_model.BAYES class
    """
    return _softmax_class.multinomial(
        PATH_SUBJECTIVE,
        S3_PATH_SUBJECTIVE,
        'subjective',
        ['negative', 'positive'],
        validate = validate,
    )


def transformer(model = 'xlnet', size = 'base', validate = True):
    """
    Load Transformer subjectivity model.

    Parameters
    ----------
    model : str, optional (default='bert')
        Model architecture supported. Allowed values:

        * ``'bert'`` - BERT architecture from google.
        * ``'xlnet'`` - XLNET architecture from google.
        * ``'albert'`` - ALBERT architecture from google.
    size : str, optional (default='base')
        Model size supported. Allowed values:

        * ``'base'`` - BASE size.
        * ``'small'`` - SMALL size.
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    MODEL : Transformer class
    """
    if not isinstance(model, str):
        raise ValueError('model must be a string')
    if not isinstance(size, str):
        raise ValueError('size must be a string')
    if not isinstance(validate, bool):
        raise ValueError('validate must be a boolean')

    model = model.lower()
    size = size.lower()
    if model not in _availability:
        raise Exception(
            'model not supported, please check supported models from malaya.subjective.available_transformer_model()'
        )
    if size not in _availability[model]:
        raise Exception(
            'size not supported, please check supported models from malaya.subjective.available_transformer_model()'
        )
    return _softmax_class.transformer(
        PATH_SUBJECTIVE,
        S3_PATH_SUBJECTIVE,
        'subjective',
        ['negative', 'positive'],
        model = model,
        size = size,
        validate = validate,
    )
